from pathlib import Path

from cg.apps.tb import TrailblazerAPI
from cg.constants import FileExtensions
from cg.constants.constants import WorkflowManager
from cg.services.analysis_starter.configurator.models.microsalt import MicrosaltCaseConfig
from cg.services.analysis_starter.submitters.subprocess.submitter import SubprocessSubmitter
from cg.services.analysis_starter.tracker.tracker import Tracker
from cg.store.models import Case
from cg.store.store import Store


class MicrosaltTracker(Tracker):
    def __init__(
        self,
        store: Store,
        subprocess_submitter: SubprocessSubmitter,
        trailblazer_api: TrailblazerAPI,
        workflow_root: str,
    ):
        super().__init__(store=store, trailblazer_api=trailblazer_api, workflow_root=workflow_root)
        self.subprocess_submitter = subprocess_submitter

    def _workflow_manager(self) -> WorkflowManager:
        return WorkflowManager.Slurm

    def _get_job_ids_path(self, case_id: str) -> Path:
        project_id: str = self._get_file_name_start(case_id)
        return Path(
            self.workflow_root,
            "results",
            "reports",
            "trailblazer",
            f"{project_id}_slurm_ids{FileExtensions.YAML}",
        )

    def _get_file_name_start(self, case_id: str) -> str:
        """Returns the LIMS project id if the case contains multiple samples, else the sample id."""
        case: Case = self.store.get_case_by_internal_id(case_id)
        sample_id: str = case.samples[0].internal_id
        if len(case.samples) == 1:
            return sample_id
        return self._extract_project_id(sample_id)

    @staticmethod
    def _extract_project_id(sample_id: str) -> str:
        return sample_id.rsplit("A", maxsplit=1)[0]

    def _get_workflow_version(self, case_config: MicrosaltCaseConfig) -> str:
        return self.subprocess_submitter.get_workflow_version(case_config)



"""Module for Gens API."""

import logging

from cg.utils import Process
from cg.utils.dict import get_list_from_dictionary

LOG = logging.getLogger(__name__)


class GensAPI:
    """API for Gens."""

    def __init__(self, config: dict[str, dict[str, str]]):
        self.binary_path: str = config["gens"]["binary_path"]
        self.config_path: str = config["gens"]["config_path"]
        self.process: Process = Process(
            binary=self.binary_path, config=self.config_path, config_parameter="--env-file"
        )
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set the dry run state."""
        self.dry_run = dry_run

    def load(
        self,
        baf_path: str,
        case_id: str,
        coverage_path: str,
        genome_build: str,
        sample_id: str,
    ) -> None:
        """Load Gens sample file paths into database."""
        load_params: dict[str, str] = {
            "--sample-id": sample_id,
            "--genome-build": genome_build,
            "--baf": baf_path,
            "--coverage": coverage_path,
            "--case-id": case_id,
        }
        load_call_params: list[str] = ["load", "sample", "--force"]
        load_call_params += get_list_from_dictionary(load_params)
        self.process.run_command(parameters=load_call_params, dry_run=self.dry_run)
        if self.process.stderr:
            LOG.warning(self.process.stderr)

    def __str__(self):
        return f"GensAPI(dry_run: {self.dry_run})"



""" class to use when the click option is an Enum
based on suggestion https://github.com/pallets/click/issues/605#issuecomment-277539425
"""

from enum import EnumMeta

import rich_click as click


class EnumChoice(click.Choice):
    """class to use when the click option is an Enum"""

    def __init__(self, enum, case_sensitive=False, use_value=True):
        if isinstance(enum, tuple):
            if use_value:
                choices = (_.value for _ in enum)
            else:
                choices = (_.name for _ in enum)
        elif isinstance(enum, EnumMeta):
            if use_value:
                choices = enum
            else:
                choices = enum.__members__
        else:
            raise TypeError("`enum` must be `tuple` or `Enum`")

        if not case_sensitive:
            choices = (_.lower() for _ in choices)

        self.__enum = enum
        self.__case_sensitive = case_sensitive
        self.__use_value = use_value

        super().__init__(list(sorted(set(choices))))

    def convert(self, value, param, ctx):
        if not self.__case_sensitive:
            value = value.lower()

        value = super().convert(value, param, ctx)

        if self.__use_value:
            if not self.__case_sensitive:
                return next(_ for _ in self.__enum if _.value.lower() == value.lower())
            else:
                return next(_ for _ in self.__enum if _.value == value)
        else:
            if not self.__case_sensitive:
                return next(_ for _ in self.__enum if _.name.lower() == value.lower())
            else:
                return next(_ for _ in self.__enum if _.name == value)



from cg.exc import CgError


class PostProcessingRunFileManagerError(CgError):
    """Error raised if something goes wrong managing the sequencing run files."""

    pass


class PostProcessingRunDataGeneratorError(CgError):
    """Error raised if something goes wrong parsing the run directory data."""

    pass


class PostProcessingParsingError(CgError):
    """Error raised if something goes wrong parsing the sequencing run metrics."""

    pass


class PostProcessingDataTransferError(CgError):
    """Error raised if something goes wrong creating the DTOs for post-processing."""

    pass


class PostProcessingStoreDataError(CgError):
    """Error raised if something goes wrong storing the post-processing data in StatusDB."""

    pass


class PostProcessingStoreFileError(CgError):
    """Error raised if something goes wrong storing the post-processing files in Housekeeper."""

    pass


class PostProcessingError(CgError):
    """Error raised if something goes wrong during post-processing."""

    pass



from typing import Annotated

from pydantic import BeforeValidator, Field

from cg.models.orders.sample_base import NAME_PATTERN, SexEnum, StatusEnum
from cg.services.orders.validation.constants import ElutionBuffer
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.utils import parse_buffer


class NalloSample(Sample):
    elution_buffer: Annotated[ElutionBuffer | None, BeforeValidator(parse_buffer)] = None
    father: str | None = Field(None, pattern=NAME_PATTERN)
    mother: str | None = Field(None, pattern=NAME_PATTERN)
    phenotype_groups: list[str] | None = None
    phenotype_terms: list[str] | None = None
    require_qc_ok: bool = False
    sex: SexEnum
    source: str
    status: StatusEnum
    subject_id: str = Field(pattern=NAME_PATTERN, max_length=128)



from flask import Blueprint, jsonify, request

from cg.server.endpoints.utils import before_request
from cg.server.ext import db
from cg.store.models import Analysis

ANALYSES_BLUEPRINT = Blueprint("analyses", __name__, url_prefix="/api/v1")
ANALYSES_BLUEPRINT.before_request(before_request)


@ANALYSES_BLUEPRINT.route("/analyses")
def get_analyses():
    """Return analyses."""
    if request.args.get("status") == "delivery":
        analyses: list[Analysis] = db.get_analyses_to_deliver_for_pipeline()
    elif request.args.get("status") == "upload":
        analyses: list[Analysis] = db.get_analyses_to_upload()
    else:
        analyses: list[Analysis] = db.get_analyses()
    parsed_analysis: list[dict] = [analysis.to_dict() for analysis in analyses[:30]]
    return jsonify(analyses=parsed_analysis, total=len(analyses))



from cg.constants.sequencing import SequencingPlatform
from cg.models.nf_analysis import NextflowSampleSheetEntry, WorkflowParameters
from cg.models.qc_metrics import QCMetrics


class TaxprofilerQCMetrics(QCMetrics):
    """Taxprofiler QC metrics."""

    after_filtering_gc_content: float
    after_filtering_read1_mean_length: float
    after_filtering_read2_mean_length: float
    after_filtering_total_reads: float
    average_length: float
    pct_duplication: float
    raw_total_sequences: float
    reads_mapped: float


class TaxprofilerParameters(WorkflowParameters):
    """Taxprofiler parameters."""


class TaxprofilerSampleSheetEntry(NextflowSampleSheetEntry):
    """Taxprofiler sample model is used when building the sample sheet."""

    instrument_platform: SequencingPlatform
    fasta: str

    @staticmethod
    def headers() -> list[str]:
        """Return sample sheet headers."""
        return [
            "sample",
            "run_accession",
            "instrument_platform",
            "fastq_1",
            "fastq_2",
            "fasta",
        ]

    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of list, where each list represents a line in the final file."""
        reformatted_content = []
        for run_accession, (forward_path, reverse_path) in enumerate(
            zip(self.fastq_forward_read_paths, self.fastq_reverse_read_paths), 1
        ):
            line = [
                self.name,
                run_accession,
                self.instrument_platform,
                forward_path,
                reverse_path,
                self.fasta,
            ]
            reformatted_content.append(line)
        return reformatted_content



import logging

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Workflow
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.configurator.configurator import Configurator
from cg.services.analysis_starter.constants import IMPLEMENTED_FASTQ_WORKFLOWS
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.input_fetcher.implementations.bam_fetcher import BamFetcher
from cg.services.analysis_starter.input_fetcher.implementations.fastq_fetcher import FastqFetcher
from cg.services.analysis_starter.input_fetcher.input_fetcher import InputFetcher
from cg.services.analysis_starter.service import AnalysisStarter
from cg.services.analysis_starter.submitters.seqera_platform.client import SeqeraPlatformClient
from cg.services.analysis_starter.submitters.seqera_platform.submitter import (
    SeqeraPlatformSubmitter,
)
from cg.services.analysis_starter.submitters.submitter import Submitter
from cg.services.analysis_starter.submitters.subprocess.submitter import SubprocessSubmitter
from cg.services.analysis_starter.tracker.implementations.microsalt import MicrosaltTracker
from cg.services.analysis_starter.tracker.implementations.mip_dna import MIPDNATracker
from cg.services.analysis_starter.tracker.implementations.nextflow import NextflowTracker
from cg.services.analysis_starter.tracker.tracker import Tracker
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class AnalysisStarterFactory:
    def __init__(self, cg_config: CGConfig):
        self.cg_config = cg_config
        self.configurator_factory = ConfiguratorFactory(cg_config)
        self.housekeeper_api: HousekeeperAPI = cg_config.housekeeper_api
        self.store: Store = cg_config.status_db

    def get_analysis_starter_for_case(self, case_id: str) -> AnalysisStarter:
        LOG.debug(f"Getting analysis starter for {case_id}")
        workflow: Workflow = self.store.get_case_workflow(case_id)
        return self.get_analysis_starter_for_workflow(workflow)

    def get_analysis_starter_for_workflow(self, workflow: Workflow) -> AnalysisStarter:
        LOG.debug(f"Getting a {workflow} analysis starter")
        configurator: Configurator = self.configurator_factory.get_configurator(workflow)
        input_fetcher: InputFetcher = self._get_input_fetcher(workflow)
        submitter: Submitter = self._get_submitter(workflow)
        tracker: Tracker = self._get_tracker(workflow)
        return AnalysisStarter(
            configurator=configurator,
            input_fetcher=input_fetcher,
            store=self.store,
            submitter=submitter,
            tracker=tracker,
            workflow=workflow,
        )

    def _get_input_fetcher(self, workflow: Workflow) -> InputFetcher:
        if workflow in IMPLEMENTED_FASTQ_WORKFLOWS:
            spring_archive_api = SpringArchiveAPI(
                status_db=self.store,
                housekeeper_api=self.housekeeper_api,
                data_flow_config=self.cg_config.data_flow,
            )
            compress_api = CompressAPI(
                hk_api=self.housekeeper_api,
                crunchy_api=self.cg_config.crunchy_api,
                demux_root=self.cg_config.run_instruments.illumina.demultiplexed_runs_dir,
            )
            return FastqFetcher(
                compress_api=compress_api,
                housekeeper_api=self.housekeeper_api,
                spring_archive_api=spring_archive_api,
                status_db=self.store,
            )
        elif workflow == Workflow.NALLO:
            return BamFetcher(housekeeper_api=self.housekeeper_api, status_db=self.store)
        raise NotImplementedError(f"No input fetcher for workflow {workflow}")

    def _get_submitter(self, workflow: Workflow) -> Submitter:
        if workflow in [
            Workflow.NALLO,
            Workflow.RAREDISEASE,
            Workflow.RNAFUSION,
            Workflow.TAXPROFILER,
        ]:
            return self._get_seqera_platform_submitter()
        else:
            return SubprocessSubmitter()

    def _get_seqera_platform_submitter(self) -> SeqeraPlatformSubmitter:
        client = SeqeraPlatformClient(config=self.cg_config.seqera_platform)
        return SeqeraPlatformSubmitter(
            client=client,
            compute_environment_ids=self.cg_config.seqera_platform.compute_environments,
        )

    def _get_tracker(self, workflow: Workflow) -> Tracker:
        if workflow in [
            Workflow.NALLO,
            Workflow.RAREDISEASE,
            Workflow.RNAFUSION,
            Workflow.TAXPROFILER,
        ]:
            return NextflowTracker(
                store=self.store,
                trailblazer_api=self.cg_config.trailblazer_api,
                workflow_root=getattr(self.cg_config, workflow).root,
            )
        elif workflow == Workflow.MICROSALT:
            return MicrosaltTracker(
                store=self.store,
                subprocess_submitter=SubprocessSubmitter(),
                trailblazer_api=self.cg_config.trailblazer_api,
                workflow_root=self.cg_config.microsalt.root,
            )
        elif workflow == Workflow.MIP_DNA:
            return MIPDNATracker(
                store=self.store,
                trailblazer_api=self.cg_config.trailblazer_api,
                workflow_root=self.cg_config.mip_rd_dna.root,
            )



import logging
from pathlib import Path

from cg.io.yaml import read_yaml, write_yaml_nextflow_style
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.abstract import (
    ParamsFileCreator,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.models import (
    RNAFusionParameters,
)
from cg.services.analysis_starter.configurator.file_creators.nextflow.params_file.utils import (
    replace_values_in_params_file,
)

LOG = logging.getLogger(__name__)


class RNAFusionParamsFileCreator(ParamsFileCreator):

    def create(self, case_id: str, file_path: Path, sample_sheet_path: Path) -> None:
        LOG.debug(f"Creating params file for case {case_id}")
        content: dict = self._get_content(
            case_run_directory=file_path.parent, sample_sheet_path=sample_sheet_path
        )
        write_yaml_nextflow_style(file_path=file_path, content=content)

    def _get_content(self, case_run_directory: Path, sample_sheet_path: Path) -> dict:
        """Return the merged dictionary with case-specific parameters and workflow parameters."""
        case_parameters = RNAFusionParameters(
            input=sample_sheet_path,
            outdir=case_run_directory,
        )
        workflow_parameters: dict = read_yaml(self.params)
        parameters: dict = case_parameters.model_dump() | workflow_parameters
        curated_parameters: dict = replace_values_in_params_file(parameters)
        return curated_parameters



from cg.services.orders.validation.rules.order.rules import (
    validate_customer_can_skip_reception_control,
    validate_customer_exists,
    validate_user_belongs_to_customer,
)

ORDER_RULES: list[callable] = [
    validate_customer_can_skip_reception_control,
    validate_customer_exists,
    validate_user_belongs_to_customer,
]



from pydantic import BaseModel, ConfigDict

from cg.models.orders.constants import OrderType
from cg.services.orders.validation.models.order import Order
from cg.services.orders.validation.order_types.balsamic.models.order import BalsamicOrder
from cg.services.orders.validation.order_types.balsamic.validation_rules import (
    BALSAMIC_CASE_RULES,
    BALSAMIC_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.balsamic_umi.models.order import BalsamicUmiOrder
from cg.services.orders.validation.order_types.balsamic_umi.validation_rules import (
    BALSAMIC_UMI_CASE_RULES,
    BALSAMIC_UMI_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.fastq.models.order import FastqOrder
from cg.services.orders.validation.order_types.fastq.validation_rules import FASTQ_SAMPLE_RULES
from cg.services.orders.validation.order_types.fluffy.models.order import FluffyOrder
from cg.services.orders.validation.order_types.fluffy.validation_rules import FLUFFY_SAMPLE_RULES
from cg.services.orders.validation.order_types.metagenome.models.order import MetagenomeOrder
from cg.services.orders.validation.order_types.metagenome.validation_rules import (
    METAGENOME_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.microbial_fastq.models.order import (
    MicrobialFastqOrder,
)
from cg.services.orders.validation.order_types.microbial_fastq.validation_rules import (
    MICROBIAL_FASTQ_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.microsalt.models.order import MicrosaltOrder
from cg.services.orders.validation.order_types.microsalt.validation_rules import (
    MICROSALT_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.mip_dna.models.order import MIPDNAOrder
from cg.services.orders.validation.order_types.mip_dna.validation_rules import (
    MIP_DNA_CASE_RULES,
    MIP_DNA_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.mip_rna.models.order import MIPRNAOrder
from cg.services.orders.validation.order_types.mip_rna.validation_rules import (
    MIP_RNA_CASE_RULES,
    MIP_RNA_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.mutant.models.order import MutantOrder
from cg.services.orders.validation.order_types.mutant.validation_rules import MUTANT_SAMPLE_RULES
from cg.services.orders.validation.order_types.nallo.models.order import NalloOrder
from cg.services.orders.validation.order_types.nallo.validation_rules import (
    NALLO_CASE_RULES,
    NALLO_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.order_validation_rules import ORDER_RULES
from cg.services.orders.validation.order_types.pacbio_long_read.models.order import PacbioOrder
from cg.services.orders.validation.order_types.pacbio_long_read.validation_rules import (
    PACBIO_LONG_READ_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.raredisease.models.order import RarediseaseOrder
from cg.services.orders.validation.order_types.raredisease.validation_rules import (
    RAREDISEASE_CASE_RULES,
    RAREDISEASE_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.rml.models.order import RMLOrder
from cg.services.orders.validation.order_types.rml.validation_rules import RML_SAMPLE_RULES
from cg.services.orders.validation.order_types.rna_fusion.models.order import RNAFusionOrder
from cg.services.orders.validation.order_types.rna_fusion.validation_rules import (
    RNAFUSION_CASE_RULES,
    RNAFUSION_CASE_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.taxprofiler.models.order import TaxprofilerOrder
from cg.services.orders.validation.order_types.taxprofiler.validation_rules import (
    TAXPROFILER_SAMPLE_RULES,
)
from cg.services.orders.validation.order_types.tomte.models.order import TomteOrder
from cg.services.orders.validation.order_types.tomte.validation_rules import (
    TOMTE_CASE_RULES,
    TOMTE_CASE_SAMPLE_RULES,
)


class RuleSet(BaseModel):
    case_rules: list[callable] = []
    case_sample_rules: list[callable] = []
    order_rules: list[callable] = ORDER_RULES
    sample_rules: list[callable] = []

    model_config = ConfigDict(arbitrary_types_allowed=True)


ORDER_TYPE_RULE_SET_MAP: dict[OrderType, RuleSet] = {
    OrderType.BALSAMIC: RuleSet(
        case_rules=BALSAMIC_CASE_RULES, case_sample_rules=BALSAMIC_CASE_SAMPLE_RULES
    ),
    OrderType.BALSAMIC_UMI: RuleSet(
        case_rules=BALSAMIC_UMI_CASE_RULES,
        case_sample_rules=BALSAMIC_UMI_CASE_SAMPLE_RULES,
    ),
    OrderType.FASTQ: RuleSet(
        sample_rules=FASTQ_SAMPLE_RULES,
    ),
    OrderType.FLUFFY: RuleSet(
        sample_rules=FLUFFY_SAMPLE_RULES,
    ),
    OrderType.METAGENOME: RuleSet(
        sample_rules=METAGENOME_SAMPLE_RULES,
    ),
    OrderType.MICROBIAL_FASTQ: RuleSet(
        sample_rules=MICROBIAL_FASTQ_SAMPLE_RULES,
    ),
    OrderType.MICROSALT: RuleSet(
        sample_rules=MICROSALT_SAMPLE_RULES,
    ),
    OrderType.MIP_DNA: RuleSet(
        case_rules=MIP_DNA_CASE_RULES, case_sample_rules=MIP_DNA_CASE_SAMPLE_RULES
    ),
    OrderType.MIP_RNA: RuleSet(
        case_rules=MIP_RNA_CASE_RULES,
        case_sample_rules=MIP_RNA_CASE_SAMPLE_RULES,
    ),
    OrderType.NALLO: RuleSet(
        case_rules=NALLO_CASE_RULES,
        case_sample_rules=NALLO_CASE_SAMPLE_RULES,
    ),
    OrderType.PACBIO_LONG_READ: RuleSet(
        sample_rules=PACBIO_LONG_READ_SAMPLE_RULES,
    ),
    OrderType.RAREDISEASE: RuleSet(
        case_rules=RAREDISEASE_CASE_RULES,
        case_sample_rules=RAREDISEASE_CASE_SAMPLE_RULES,
    ),
    OrderType.RML: RuleSet(
        sample_rules=RML_SAMPLE_RULES,
    ),
    OrderType.RNAFUSION: RuleSet(
        case_rules=RNAFUSION_CASE_RULES,
        case_sample_rules=RNAFUSION_CASE_SAMPLE_RULES,
    ),
    OrderType.SARS_COV_2: RuleSet(
        sample_rules=MUTANT_SAMPLE_RULES,
    ),
    OrderType.TAXPROFILER: RuleSet(
        sample_rules=TAXPROFILER_SAMPLE_RULES,
    ),
    OrderType.TOMTE: RuleSet(
        case_rules=TOMTE_CASE_RULES,
        case_sample_rules=TOMTE_CASE_SAMPLE_RULES,
    ),
}

ORDER_TYPE_MODEL_MAP: dict[OrderType, type[Order]] = {
    OrderType.BALSAMIC: BalsamicOrder,
    OrderType.BALSAMIC_UMI: BalsamicUmiOrder,
    OrderType.FASTQ: FastqOrder,
    OrderType.FLUFFY: FluffyOrder,
    OrderType.METAGENOME: MetagenomeOrder,
    OrderType.MICROBIAL_FASTQ: MicrobialFastqOrder,
    OrderType.MICROSALT: MicrosaltOrder,
    OrderType.MIP_DNA: MIPDNAOrder,
    OrderType.MIP_RNA: MIPRNAOrder,
    OrderType.NALLO: NalloOrder,
    OrderType.PACBIO_LONG_READ: PacbioOrder,
    OrderType.RAREDISEASE: RarediseaseOrder,
    OrderType.RML: RMLOrder,
    OrderType.RNAFUSION: RNAFusionOrder,
    OrderType.SARS_COV_2: MutantOrder,
    OrderType.TAXPROFILER: TaxprofilerOrder,
    OrderType.TOMTE: TomteOrder,
}



from cg.constants.pedigree import Pedigree
from cg.services.orders.validation.errors.case_sample_errors import (
    DescendantAsFatherError,
    DescendantAsMotherError,
    PedigreeError,
    SampleIsOwnFatherError,
    SampleIsOwnMotherError,
)
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample
from cg.services.orders.validation.order_types.tomte.models.sample import TomteSample
from cg.services.orders.validation.rules.case_sample.pedigree.models import FamilyTree, Node


def validate_tree(pedigree: FamilyTree) -> list[PedigreeError]:
    """This performs a DFS algorithm on the family tree to find any cycles, which indicates an
    order error."""
    errors: list[PedigreeError] = []
    for node in pedigree.nodes:
        if not node.visited:
            detect_cycles(node=node, errors=errors)
    return errors


def detect_cycles(node: Node, errors: list[PedigreeError]) -> None:
    """Detect cycles in the pedigree graph using depth-first search. If a cycle is detected,
    this is considered an error."""
    node.visited = True
    node.in_current_path = True

    parents: dict[str, Node] = {Pedigree.MOTHER: node.mother, Pedigree.FATHER: node.father}

    for parent_type, parent in parents.items():
        if parent and parent.in_current_path:
            error: PedigreeError = get_error(node=node, parent_type=parent_type)
            errors.append(error)
        elif parent and not parent.visited:
            detect_cycles(node=parent, errors=errors)
    node.in_current_path = False


def get_error(node: Node, parent_type: str) -> PedigreeError:
    if parent_type == Pedigree.MOTHER:
        return get_mother_error(node)
    if parent_type == Pedigree.FATHER:
        return get_father_error(node)


def get_mother_error(node: Node) -> PedigreeError:
    """Called when the node's 'mother' creates a cycle in the family tree. For clearer feedback
    we distinguish between the sample being its own mother, and other more complex situations."""
    sample: TomteSample | MIPDNASample | ExistingSample = node.sample
    if node.sample_name == sample.mother:
        return SampleIsOwnMotherError(sample_index=node.sample_index, case_index=node.case_index)
    return DescendantAsMotherError(sample_index=node.sample_index, case_index=node.case_index)


def get_father_error(node: Node) -> PedigreeError:
    """Called when the node's 'father' creates a cycle in the family tree. For clearer feedback
    we distinguish between the sample being its own father, and other more complex situations."""
    sample: TomteSample = node.sample
    if node.sample_name == sample.father:
        return SampleIsOwnFatherError(sample_index=node.sample_index, case_index=node.case_index)
    return DescendantAsFatherError(sample_index=node.sample_index, case_index=node.case_index)



"""CLI support to create config and/or start BALSAMIC."""

import logging
import traceback

import rich_click as click
from pydantic.v1 import ValidationError

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.balsamic.options import (
    OPTION_CACHE_VERSION,
    OPTION_GENDER,
    OPTION_GENOME_VERSION,
    OPTION_OBSERVATIONS,
    OPTION_PANEL_BED,
    OPTION_PON_CNN,
    OPTION_QOS,
    OPTION_WORKFLOW_PROFILE,
)
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, link, resolve_compression
from cg.cli.workflow.utils import validate_force_store_option
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.cli_options import COMMENT, DRY_RUN, FORCE, LIMIT
from cg.exc import AnalysisNotReadyError, CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def balsamic(context: click.Context):
    """Cancer analysis workflow"""
    AnalysisAPI.get_help(context)

    config = context.obj
    context.obj.meta_apis["analysis_api"] = BalsamicAnalysisAPI(config=config)


balsamic.add_command(resolve_compression)
balsamic.add_command(link)


@balsamic.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_GENDER
@OPTION_GENOME_VERSION
@OPTION_PANEL_BED
@OPTION_PON_CNN
@OPTION_OBSERVATIONS
@OPTION_CACHE_VERSION
@DRY_RUN
@click.pass_obj
def config_case(
    context: CGConfig,
    case_id: str,
    gender: str,
    genome_version: str,
    panel_bed: str,
    pon_cnn: click.Path,
    observations: list[click.Path],
    cache_version: str,
    dry_run: bool,
):
    """Create config file for BALSAMIC analysis for a given CASE_ID."""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    try:
        LOG.info(f"Creating config file for {case_id}.")
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.config_case(
            case_id=case_id,
            gender=gender,
            genome_version=genome_version,
            panel_bed=panel_bed,
            pon_cnn=pon_cnn,
            observations=observations,
            cache_version=cache_version,
            dry_run=dry_run,
        )
    except CgError as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not create config: {error_info}")
        raise click.Abort()
    except Exception as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not create config: {error_info}")
        raise click.Abort()


@balsamic.command("run")
@ARGUMENT_CASE_ID
@OPTION_WORKFLOW_PROFILE
@DRY_RUN
@OPTION_QOS
@click.pass_obj
def run(
    context: CGConfig,
    case_id: str,
    workflow_profile: click.Path,
    slurm_quality_of_service: str,
    dry_run: bool,
):
    """Run balsamic analysis for given CASE ID"""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.check_analysis_ongoing(case_id)
        analysis_api.run_analysis(
            case_id=case_id,
            workflow_profile=workflow_profile,
            slurm_quality_of_service=slurm_quality_of_service,
            dry_run=dry_run,
        )
        if dry_run:
            return
        analysis_api.on_analysis_started(case_id)
    except Exception as error:
        error_info = f"Error: {type(error).__name__}: {str(error)}\n{traceback.format_exc()}"
        LOG.error(f"Could not run analysis: {error_info}")
        raise click.Abort()


@balsamic.command("report-deliver")
@ARGUMENT_CASE_ID
@DRY_RUN
@FORCE
@click.pass_obj
def report_deliver(context: CGConfig, case_id: str, dry_run: bool, force: bool):
    """Create a Housekeeper deliverables file for a given case ID."""
    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    try:
        analysis_api.status_db.verify_case_exists(case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.trailblazer_api.verify_latest_analysis_is_completed(
            case_id=case_id, force=force
        )
        analysis_api.report_deliver(case_id=case_id, dry_run=dry_run)
    except CgError as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()
    except Exception as error:
        LOG.error(f"Could not create report file: {error}")
        raise click.Abort()


@balsamic.command("store-housekeeper")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_obj
def store_housekeeper(
    context: CGConfig, case_id: str, comment: str | None, dry_run: bool, force: bool
):
    """Store a finished analysis in Housekeeper and StatusDB."""

    analysis_api: AnalysisAPI = context.meta_apis["analysis_api"]
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.verify_case_config_file_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.verify_deliverables_file_exists(case_id=case_id)
        _, version = analysis_api.create_housekeeper_bundle(
            case_id=case_id, dry_run=dry_run, force=force
        )
        analysis_api.update_analysis_as_completed_statusdb(
            case_id=case_id, hk_version_id=version.id, comment=comment, dry_run=dry_run, force=force
        )
        analysis_api.set_statusdb_action(case_id=case_id, action=None, dry_run=dry_run)
    except ValidationError as error:
        LOG.warning("Deliverables file is malformed")
        raise error
    except CgError as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}")
        raise click.Abort()
    except Exception as error:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {error}!")
        housekeeper_api.rollback()
        status_db.session.rollback()
        raise click.Abort()


@balsamic.command("start")
@ARGUMENT_CASE_ID
@OPTION_GENDER
@OPTION_GENOME_VERSION
@OPTION_QOS
@DRY_RUN
@OPTION_PANEL_BED
@OPTION_PON_CNN
@OPTION_CACHE_VERSION
@OPTION_OBSERVATIONS
@OPTION_WORKFLOW_PROFILE
@click.pass_context
def start(
    context: click.Context,
    case_id: str,
    gender: str,
    genome_version: str,
    cache_version: str,
    panel_bed: str,
    pon_cnn: str,
    observations: list[click.Path],
    slurm_quality_of_service: str,
    workflow_profile: click.Path,
    dry_run: bool,
):
    """Start full workflow for case ID."""
    analysis_api: BalsamicAnalysisAPI = context.obj.meta_apis["analysis_api"]
    analysis_api.prepare_fastq_files(case_id=case_id, dry_run=dry_run)
    LOG.info(f"Starting analysis for {case_id}")
    context.invoke(link, case_id=case_id, dry_run=dry_run)
    context.invoke(
        config_case,
        case_id=case_id,
        gender=gender,
        genome_version=genome_version,
        cache_version=cache_version,
        panel_bed=panel_bed,
        pon_cnn=pon_cnn,
        observations=observations,
        dry_run=dry_run,
    )
    context.invoke(
        run,
        case_id=case_id,
        workflow_profile=workflow_profile,
        slurm_quality_of_service=slurm_quality_of_service,
        dry_run=dry_run,
    )


@balsamic.command("start-available")
@DRY_RUN
@LIMIT
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False, limit: int | None = None):
    """Start full workflow for all cases ready for analysis"""

    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    cases: list[Case] = analysis_api.get_cases_to_analyze(limit=limit)
    LOG.info(f"Starting {len(cases)} available Balsamic cases")

    exit_code: int = EXIT_SUCCESS
    for case in cases:
        try:
            context.invoke(start, case_id=case.internal_id, dry_run=dry_run)
        except AnalysisNotReadyError as error:
            LOG.error(error)
        except CgError as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
        except Exception as error:
            LOG.error(f"Unspecified error occurred: {error}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort


@balsamic.command("store")
@ARGUMENT_CASE_ID
@COMMENT
@DRY_RUN
@FORCE
@click.pass_context
def store(context: click.Context, case_id: str, comment: str | None, dry_run: bool, force: bool):
    """Generate Housekeeper report for CASE ID and store in Housekeeper"""
    LOG.info(f"Storing analysis for {case_id}")
    validate_force_store_option(force=force, comment=comment)
    context.invoke(report_deliver, case_id=case_id, dry_run=dry_run, force=force)
    context.invoke(store_housekeeper, case_id=case_id, comment=comment, force=force)


@balsamic.command("store-available")
@DRY_RUN
@click.pass_context
def store_available(context: click.Context, dry_run: bool) -> None:
    """Store bundles for all finished analyses in Housekeeper"""

    analysis_api: AnalysisAPI = context.obj.meta_apis["analysis_api"]

    was_successful: bool = True
    for case_obj in analysis_api.get_cases_to_store():
        LOG.info(f"Storing deliverables for {case_obj.internal_id}")
        try:
            context.invoke(store, case_id=case_obj.internal_id, dry_run=dry_run)
        except Exception as exception_object:
            LOG.error(f"Error storing {case_obj.internal_id}: {exception_object}")
            was_successful = False
    if not was_successful:
        raise click.Abort()



"""Utility functions for the upload cli commands."""

import logging

import rich_click as click

from cg.apps.scout.scoutapi import ScoutAPI
from cg.constants import Workflow
from cg.constants.constants import MAX_ITEMS_TO_RETRIEVE
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis
from cg.store.store import Store

LOG = logging.getLogger(__name__)


def suggest_cases_to_upload(status_db: Store, workflow: Workflow | None = None) -> None:
    """Print a list of suggested cases to upload."""
    LOG.warning("Provide a case, suggestions:")
    records: list[Analysis] = status_db.get_analyses_to_upload(workflow=workflow)[
        :MAX_ITEMS_TO_RETRIEVE
    ]
    for case_obj in records:
        click.echo(case_obj)


def get_scout_api(cg_config: CGConfig, case_id: str) -> ScoutAPI:
    workflow = cg_config.status_db.get_case_by_internal_id(case_id).data_analysis
    return cg_config.scout_api_38 if workflow == Workflow.NALLO else cg_config.scout_api_37



from typing import Iterator

from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.creator import (
    NextflowFastqSampleSheetCreator,
)
from cg.store.models import Case, CaseSample

HEADERS: list[str] = [
    "sample",
    "lane",
    "fastq_1",
    "fastq_2",
    "sex",
    "phenotype",
    "paternal_id",
    "maternal_id",
    "case_id",
]


class RarediseaseSampleSheetCreator(NextflowFastqSampleSheetCreator):

    def _get_content(self, case_id: str) -> list[list[str]]:
        """Return formatted information required to build a sample sheet for a raredisease case.
        This contains information for all samples linked to the case."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content: list[list[str]] = [HEADERS]
        for link in case.links:
            sample_sheet_content.extend(self._get_sample_sheet_content_per_sample(case_sample=link))
        return sample_sheet_content

    def _get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        content: list[list[str]] = []
        paired_fastq_paths: Iterator[tuple[str, str]] = self._get_paired_read_paths(
            case_sample.sample
        )
        for incremental_id, (fastq_forward_read_path, fastq_reverse_read_path) in enumerate(
            paired_fastq_paths, start=1
        ):
            content.append(
                [
                    case_sample.sample.internal_id,
                    incremental_id,
                    fastq_forward_read_path,
                    fastq_reverse_read_path,
                    str(self._get_sex_code(case_sample.sample.sex)),
                    str(self._get_phenotype_code(case_sample.status)),
                    case_sample.get_paternal_sample_id,
                    case_sample.get_maternal_sample_id,
                    case_sample.case.internal_id,
                ]
            )
        return content

    @staticmethod
    def _get_phenotype_code(phenotype: str) -> int:
        """Return Raredisease phenotype code."""
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def _get_sex_code(sex: str) -> PlinkSex:
        """Return Raredisease sex code."""
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code



from collections.abc import Iterator

from cg.constants.sequencing import SequencingPlatform
from cg.constants.symbols import EMPTY_STRING
from cg.services.analysis_starter.configurator.file_creators.nextflow.sample_sheet.creator import (
    NextflowFastqSampleSheetCreator,
)
from cg.store.models import Case, Sample

HEADERS: list[str] = [
    "sample",
    "run_accession",
    "instrument_platform",
    "fastq_1",
    "fastq_2",
    "fasta",
]


class TaxprofilerSampleSheetCreator(NextflowFastqSampleSheetCreator):

    def _get_content(self, case_id: str) -> list[list[str]]:
        """Return formatted information required to build a sample sheet for a case.
        This contains information for all samples linked to the case."""
        case: Case = self.store.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content: list[list[str]] = [HEADERS]
        for sample in case.samples:
            sample_sheet_content.extend(self._get_sample_sheet_content_per_sample(sample))
        return sample_sheet_content

    def _get_sample_sheet_content_per_sample(self, sample: Sample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        content: list[list[str]] = []
        paired_fastq_files: Iterator[tuple[str, str]] = self._get_paired_read_paths(sample)
        for incremental_id, (forward_file, reverse_file) in enumerate(paired_fastq_files, start=1):
            entry: list[str] = [
                sample.name,
                incremental_id,
                SequencingPlatform.ILLUMINA,
                forward_file,
                reverse_file,
                EMPTY_STRING,
            ]
            content.append(entry)
        return content



"""MIP-RNA upload API."""

import logging
from subprocess import CalledProcessError

import rich_click as click

from cg.cli.upload.scout import upload_rna_to_scout
from cg.constants import DataDelivery
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Case

LOG = logging.getLogger(__name__)


class MipRNAUploadAPI(UploadAPI):
    """MIP-RNA upload API."""

    def __init__(self, config: CGConfig):
        self.analysis_api: MipRNAAnalysisAPI = MipRNAAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Case, restart: bool) -> None:
        """Uploads MIP-RNA analysis data and files."""
        analysis: Analysis = self.status_db.get_latest_completed_analysis_for_case(case.internal_id)
        self.update_upload_started_at(analysis=analysis)

        self.upload_files_to_customer_inbox(case=case)

        # Scout specific upload
        if DataDelivery.SCOUT in case.data_delivery:
            try:
                ctx.invoke(upload_rna_to_scout, case_id=case.internal_id)
            except CalledProcessError as error:
                LOG.error(error)
                return
        self.update_uploaded_at(analysis)



from abc import ABC, abstractmethod
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import DeliveryFiles


class DestinationFilesMover(ABC):
    @abstractmethod
    def move_files(self, delivery_files: DeliveryFiles, delivery_base_path: Path) -> DeliveryFiles:
        """Move files to the delivery folder."""
        pass



from pathlib import Path

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import FileExtensions, SPACE
from cg.constants.encryption import GPGParameters
from cg.constants.priority import SlurmQos
from cg.exc import FlowCellError, IlluminaRunEncryptionError
from cg.meta.encryption.encryption import EncryptionAPI, LIMIT_PIGZ_TASK, LOG
from cg.meta.encryption.sbatch import ILLUMINA_RUN_ENCRYPT_ERROR, ILLUMINA_RUN_ENCRYPT_COMMANDS
from cg.meta.tar.tar import TarAPI
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.models.slurm.sbatch import Sbatch


class IlluminaRunEncryptionService(EncryptionAPI):
    """Encryption functionality for Illumina sequencing runs."""

    def __init__(
        self,
        binary_path: str,
        encryption_dir: Path,
        run_dir_data: IlluminaRunDirectoryData,
        pigz_binary_path: str,
        sbatch_parameter: dict[str, str | int],
        slurm_api: SlurmAPI,
        tar_api: TarAPI,
        dry_run: bool = False,
    ):
        super().__init__(binary_path=binary_path, dry_run=dry_run)
        self.run_dir_data: IlluminaRunDirectoryData = run_dir_data
        self.encryption_dir: Path = encryption_dir
        self.tar_api: TarAPI = tar_api
        self.pigz_binary_path: str = pigz_binary_path
        self.slurm_api: SlurmAPI = slurm_api
        self.slurm_account: str = sbatch_parameter.get("account")
        self.slurm_hours: int = sbatch_parameter.get("hours")
        self.slurm_mail_user: str = sbatch_parameter.get("mail_user")
        self.slurm_memory: int = sbatch_parameter.get("memory")
        self.slurm_number_tasks: int = sbatch_parameter.get("number_tasks")
        self.slurm_api.set_dry_run(dry_run=self.dry_run)

    @property
    def run_encryption_dir(self) -> Path:
        return Path(self.encryption_dir, self.run_dir_data.full_name)

    @property
    def run_encrypt_tmp_dir(self) -> Path:
        return Path(self.run_encryption_dir, "tmp")

    @property
    def run_encrypt_file_path_prefix(self) -> Path:
        return Path(self.run_encryption_dir, self.run_dir_data.id)

    @property
    def pending_file_path(self) -> Path:
        return Path(self.run_encrypt_file_path_prefix.with_suffix(FileExtensions.PENDING))

    @property
    def complete_file_path(self) -> Path:
        return Path(self.run_encrypt_file_path_prefix.with_suffix(FileExtensions.COMPLETE))

    @property
    def encrypted_gpg_file_path(self) -> Path:
        return Path(
            self.run_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}"
            )
        )

    @property
    def encrypted_md5sum_file_path(self) -> Path:
        return Path(
            self.run_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.MD5SUM}"
            )
        )

    @property
    def decrypted_md5sum_file_path(self) -> Path:
        return Path(
            self.run_encrypt_file_path_prefix.with_suffix(
                f"{FileExtensions.TAR}{FileExtensions.GZIP}.degpg{FileExtensions.MD5SUM}"
            )
        )

    @property
    def symmetric_passphrase_file_path(self) -> Path:
        return Path(self.run_encrypt_file_path_prefix.with_suffix(FileExtensions.PASS_PHRASE))

    @property
    def final_passphrase_file_path(self) -> Path:
        return Path(self.run_encrypt_file_path_prefix.with_suffix(f".key{FileExtensions.GPG}"))

    def get_run_symmetric_encryption_command(
        self, output_file: Path, passphrase_file_path: Path
    ) -> str:
        """Generates the Gpg command for symmetric encryption of file."""
        encryption_parameters: list[str] = (
            [self.binary_path]
            + GPGParameters.SYMMETRIC_ENCRYPTION
            + [passphrase_file_path.as_posix()]
            + GPGParameters.OUTPUT_PARAMETER
            + [output_file.as_posix()]
        )
        return SPACE.join(encryption_parameters)

    def get_run_symmetric_decryption_command(
        self, input_file: Path, passphrase_file_path: Path
    ) -> str:
        """Generates the Gpg command for symmetric decryption."""
        decryption_parameters: list[str] = (
            [self.binary_path]
            + GPGParameters.SYMMETRIC_DECRYPTION
            + [passphrase_file_path.as_posix(), input_file.as_posix()]
        )
        return SPACE.join(decryption_parameters)

    def is_encryption_possible(self) -> bool | None:
        """Check if requirements for encryption are meet.
        Raises:
            FlowCellError if sequencing is not ready, encryption is pending or complete.
        """
        if not self.run_dir_data.is_sequencing_run_ready():
            raise FlowCellError(f"Run: {self.run_dir_data.id} is not ready")
        if self.complete_file_path.exists():
            raise IlluminaRunEncryptionError(
                f"Encryption already completed for run: {self.run_dir_data.id}"
            )
        if self.pending_file_path.exists():
            raise IlluminaRunEncryptionError(
                f"Encryption already started for run: {self.run_dir_data.id}"
            )
        return True

    def make_tmp_encrypt_dir(self) -> str:
        return f"mkdir -p {self.run_encrypt_tmp_dir}"

    def copy_run_dir_to_tmp(self) -> str:
        return f"cp -r {self.run_dir_data.path} {self.run_encrypt_tmp_dir}"

    def remove_tmp_encrypt_dir(self) -> str:
        return f"rm -rf {self.run_encrypt_tmp_dir}"

    def encrypt_run(
        self,
    ) -> None:
        """Encrypt a sequencing run via GPG and SLURM."""
        error_function: str = ILLUMINA_RUN_ENCRYPT_ERROR.format(
            pending_file_path=self.pending_file_path
        )
        commands: str = ILLUMINA_RUN_ENCRYPT_COMMANDS.format(
            symmetric_passphrase=self.get_symmetric_passphrase_cmd(
                passphrase_file_path=self.symmetric_passphrase_file_path
            ),
            asymmetrically_encrypt_passphrase=self.get_asymmetrically_encrypt_passphrase_cmd(
                passphrase_file_path=self.symmetric_passphrase_file_path
            ),
            make_tmp_encrypt_dir=self.make_tmp_encrypt_dir(),
            copy_run_dir_to_tmp=self.copy_run_dir_to_tmp(),
            tar_encrypt_tmp_dir=self.tar_api.get_compress_cmd(input_path=self.run_encrypt_tmp_dir),
            parallel_gzip=f"{self.pigz_binary_path} -p {self.slurm_number_tasks - LIMIT_PIGZ_TASK} --fast -c",
            tee=f"tee >(md5sum > {self.encrypted_md5sum_file_path})",
            run_symmetric_encryption=self.get_run_symmetric_encryption_command(
                output_file=self.encrypted_gpg_file_path,
                passphrase_file_path=self.symmetric_passphrase_file_path,
            ),
            run_symmetric_decryption=self.get_run_symmetric_decryption_command(
                input_file=self.encrypted_gpg_file_path,
                passphrase_file_path=self.symmetric_passphrase_file_path,
            ),
            md5sum=f"md5sum > {self.decrypted_md5sum_file_path}",
            diff=f"diff -q {self.encrypted_md5sum_file_path} {self.decrypted_md5sum_file_path}",
            mv_passphrase_file=f"mv {self.symmetric_passphrase_file_path.with_suffix(FileExtensions.GPG)} {self.final_passphrase_file_path}",
            remove_pending_file=f"rm -f {self.pending_file_path}",
            flag_as_complete=f"touch {self.complete_file_path}",
            remove_tmp_encrypt_dir=self.remove_tmp_encrypt_dir(),
        )
        sbatch_parameters = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([self.run_dir_data.id, "flow_cell_encryption"]),
            log_dir=self.run_encryption_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.HIGH,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(sbatch_parameters)
        sbatch_path = Path(self.run_encrypt_file_path_prefix.with_suffix(FileExtensions.SBATCH))
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Run encryption running as job {sbatch_number}")

    def start_encryption(self) -> None:
        """Start encryption if requirements are met."""
        self.is_encryption_possible()
        self.run_encryption_dir.mkdir(exist_ok=True, parents=True)
        self.create_pending_file(pending_path=self.pending_file_path)
        self.encrypt_run()



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import (
    REMINDER_TO_DOWNLOAD_MESSAGE,
    get_caesar_delivery_path,
)
from cg.store.models import Case


class BamMessage(DeliveryMessage):

    def create_message(self, cases: list[Case]) -> str:
        delivery_path: str = get_caesar_delivery_path(cases[0])
        return (
            f"Hello,\n\n"
            f"The following samples in this order have been sequenced:\n\n"
            f"{self.generate_sample_name_list(cases)}\n\n"
            f"The bam files have been uploaded to your inbox on Caesar at:\n\n"
            f"{delivery_path} \n\n"
            f"{REMINDER_TO_DOWNLOAD_MESSAGE}"
        )

    @staticmethod
    def generate_sample_name_list(cases: list[Case]) -> str:
        return "\n".join([sample.name for case in cases for sample in case.samples])



from pydantic import AfterValidator, BaseModel
from typing_extensions import Annotated

from cg.models.workflow.validators import get_first_word


class MutantSampleConfig(BaseModel):
    CG_ID_project: str
    CG_ID_sample: str
    case_ID: str
    region_code: Annotated[str, AfterValidator(get_first_word)]
    lab_code: Annotated[str, AfterValidator(get_first_word)]
    priority: str
    Customer_ID_project: int
    Customer_ID_sample: str
    customer_id: str
    application_tag: str
    date_arrival: str
    date_libprep: str
    date_sequencing: str
    method_libprep: str
    method_sequencing: str
    sequencing_qc_pass: bool
    selection_criteria: str
    primer: str



"""
    Module for interacting with crunchy to perform:
        1. Compressing: FASTQ to SPRING
        2. Decompressing: SPRING to FASTQ
    along with the helper methods.
"""

import logging
from pathlib import Path

from cg.apps.crunchy import files
from cg.apps.crunchy.models import CrunchyFile, CrunchyMetadata
from cg.apps.crunchy.sbatch import (
    FASTQ_TO_SPRING_COMMANDS,
    FASTQ_TO_SPRING_ERROR,
    SPRING_TO_FASTQ_COMMANDS,
    SPRING_TO_FASTQ_ERROR,
)
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants.priority import SlurmQos
from cg.models.compression_data import CompressionData
from cg.models.slurm.sbatch import Sbatch

LOG = logging.getLogger(__name__)


class CrunchyAPI:
    """
    API for Crunchy.
    """

    def __init__(self, config: dict):
        self.conda_binary: str | None = config["crunchy"]["conda_binary"] or None
        self.crunchy_env: str = config["crunchy"]["slurm"]["conda_env"]
        self.dry_run: bool = False
        self.reference_path: str = config["crunchy"]["cram_reference"]
        self.slurm_api: SlurmAPI = SlurmAPI()
        self.slurm_account: str = config["crunchy"]["slurm"]["account"]
        self.slurm_hours: int = config["crunchy"]["slurm"]["hours"]
        self.slurm_mail_user: str = config["crunchy"]["slurm"]["mail_user"]
        self.slurm_memory: int = config["crunchy"]["slurm"]["memory"]
        self.slurm_number_tasks: int = config["crunchy"]["slurm"]["number_tasks"]

    def set_dry_run(self, dry_run: bool) -> None:
        """Update dry run."""
        self.dry_run = dry_run
        self.slurm_api.set_dry_run(dry_run=dry_run)

    @staticmethod
    def create_pending_file(pending_path: Path, dry_run: bool) -> None:
        """Create a pending flag file."""
        LOG.info(f"Creating pending flag {pending_path}")
        if dry_run:
            return
        pending_path.touch(exist_ok=False)

    # These are the compression/decompression methods
    def fastq_to_spring(self, compression_obj: CompressionData, sample_id: str = "") -> int:
        """Compress FASTQ files into SPRING by sending to sbatch SLURM."""
        CrunchyAPI.create_pending_file(
            pending_path=compression_obj.pending_path, dry_run=self.dry_run
        )
        log_dir: Path = files.get_log_dir(compression_obj.spring_path)
        # Generate the error function
        error_function = FASTQ_TO_SPRING_ERROR.format(
            spring_path=compression_obj.spring_path, pending_path=compression_obj.pending_path
        )
        # Generate the commands
        sbatch_parameters: Sbatch
        commands = FASTQ_TO_SPRING_COMMANDS.format(
            conda_run=f"{self.conda_binary} run --name {self.crunchy_env}",
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            pending_path=compression_obj.pending_path,
            spring_path=compression_obj.spring_path,
            tmp_dir=files.get_tmp_dir(
                prefix="spring_",
                suffix="_compress",
                base=compression_obj.analysis_dir.as_posix(),
            ),
        )
        sbatch_parameters: Sbatch = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([sample_id, compression_obj.run_name, "fastq_to_spring"]),
            log_dir=log_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.MAINTENANCE,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = files.get_fastq_to_spring_sbatch_path(
            log_dir=log_dir, run_name=compression_obj.run_name
        )
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Fastq compression running as job {sbatch_number}")
        return sbatch_number

    def spring_to_fastq(self, compression_obj: CompressionData, sample_id: str = "") -> int:
        """Decompress SPRING into FASTQ by submitting sbatch script to SLURM."""
        CrunchyAPI.create_pending_file(
            pending_path=compression_obj.pending_path, dry_run=self.dry_run
        )
        # Fetch the metadata information from a spring metadata file
        crunchy_metadata: CrunchyMetadata = files.get_crunchy_metadata(
            compression_obj.spring_metadata_path
        )
        files_info: dict[str, CrunchyFile] = files.get_spring_archive_files(crunchy_metadata)
        log_dir = files.get_log_dir(compression_obj.spring_path)

        error_function = SPRING_TO_FASTQ_ERROR.format(
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            pending_path=compression_obj.pending_path,
        )
        commands = SPRING_TO_FASTQ_COMMANDS.format(
            conda_run=f"{self.conda_binary} run --name {self.crunchy_env}",
            tmp_dir=files.get_tmp_dir(
                prefix="spring_",
                suffix="_decompress",
                base=compression_obj.analysis_dir.as_posix(),
            ),
            fastq_first=compression_obj.fastq_first,
            fastq_second=compression_obj.fastq_second,
            spring_path=compression_obj.spring_path,
            pending_path=compression_obj.pending_path,
            checksum_first=files_info["fastq_first"].checksum,
            checksum_second=files_info["fastq_second"].checksum,
        )
        sbatch_parameters: Sbatch = Sbatch(
            account=self.slurm_account,
            commands=commands,
            email=self.slurm_mail_user,
            error=error_function,
            hours=self.slurm_hours,
            job_name="_".join([sample_id, compression_obj.run_name, "spring_to_fastq"]),
            log_dir=log_dir.as_posix(),
            memory=self.slurm_memory,
            number_tasks=self.slurm_number_tasks,
            quality_of_service=SlurmQos.HIGH,
        )
        sbatch_content: str = self.slurm_api.generate_sbatch_content(sbatch_parameters)
        sbatch_path = files.get_spring_to_fastq_sbatch_path(
            log_dir=log_dir, run_name=compression_obj.run_name
        )
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info(f"Spring decompression running as job {sbatch_number}")
        return sbatch_number



from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.clients.chanjo2.client import Chanjo2APIClient
from cg.meta.backup.backup import SpringBackupAPI
from cg.meta.compress import CompressAPI
from cg.meta.delivery.delivery import DeliveryAPI
from cg.meta.encryption.encryption import SpringEncryptionAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.services.pdc_service.pdc_service import PdcService
from cg.store.store import Store


class MetaAPI:
    """MetaAPI class initializing all App APIs used within CG in non-conflicting manner"""

    def __init__(self, config: CGConfig):
        self.chanjo_api: ChanjoAPI = config.chanjo_api
        self.chanjo2_api: Chanjo2APIClient = config.chanjo2_api
        self.config = config
        self.crunchy_api: CrunchyAPI = config.crunchy_api
        self.delivery_api: DeliveryAPI = config.delivery_api
        self.genotype_api: GenotypeAPI = config.genotype_api
        self.hermes_api: HermesApi = config.hermes_api
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.madeline_api: MadelineAPI = config.madeline_api
        self.mutacc_auto_api: MutaccAutoAPI = config.mutacc_auto_api
        self.prepare_fastq_api: PrepareFastqAPI = PrepareFastqAPI(
            store=config.status_db,
            compress_api=CompressAPI(
                hk_api=config.housekeeper_api,
                crunchy_api=config.crunchy_api,
                demux_root=config.run_instruments.illumina.demultiplexed_runs_dir,
                backup_api=SpringBackupAPI(
                    encryption_api=SpringEncryptionAPI(
                        binary_path=config.dict()["encryption"]["binary_path"]
                    ),
                    hk_api=config.housekeeper_api,
                    pdc_service=PdcService(config.dict()["pdc"]["binary_path"]),
                ),
            ),
        )
        self.scout_api_37: ScoutAPI = config.scout_api_37
        self.scout_api_38: ScoutAPI = config.scout_api_38
        self.status_db: Store = config.status_db
        self.trailblazer_api: TrailblazerAPI = config.trailblazer_api



"""Module for the Pacbio database store service."""

import logging
from datetime import datetime

from cg.services.run_devices.abstract_classes import PostProcessingStoreService
from cg.services.run_devices.error_handler import handle_post_processing_errors
from cg.services.run_devices.exc import (
    PostProcessingDataTransferError,
    PostProcessingStoreDataError,
)
from cg.services.run_devices.pacbio.data_transfer_service.data_transfer_service import (
    PacBioDataTransferService,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioDTOs,
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
)
from cg.services.run_devices.pacbio.run_data_generator.run_data import PacBioRunData
from cg.store.models import PacbioSequencingRun, PacbioSMRTCell
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class PacBioStoreService(PostProcessingStoreService):
    def __init__(self, store: Store, data_transfer_service: PacBioDataTransferService):
        self.store = store
        self.data_transfer_service = data_transfer_service

    def _create_run_device(self, run_device_dto: PacBioSMRTCellDTO) -> PacbioSMRTCell:
        return self.store.create_pac_bio_smrt_cell(run_device_dto)

    def _create_instrument_run(
        self, instrument_run_dto: PacBioSequencingRunDTO, smrt_cell: PacbioSMRTCell
    ) -> PacbioSequencingRun:
        return self.store.create_pac_bio_sequencing_run(
            sequencing_run_dto=instrument_run_dto, smrt_cell=smrt_cell
        )

    def _create_sample_run_metrics(
        self,
        sample_run_metrics_dtos: list[PacBioSampleSequencingMetricsDTO],
        sequencing_run: PacbioSequencingRun,
    ) -> None:
        for sample_run_metric in sample_run_metrics_dtos:
            self.store.create_pac_bio_sample_sequencing_run(
                sample_run_metrics_dto=sample_run_metric, sequencing_run=sequencing_run
            )

    def _update_sample(
        self,
        sample_run_metrics_dtos: list[PacBioSampleSequencingMetricsDTO],
        sequencing_date: datetime,
    ) -> None:
        """Update the reads and last sequenced date for the SMRT cell samples."""
        for sample_dto in sample_run_metrics_dtos:
            self.store.update_sample_reads(
                internal_id=sample_dto.sample_internal_id, reads=sample_dto.hifi_reads
            )
            self.store.update_sample_sequenced_at(
                internal_id=sample_dto.sample_internal_id, date=sequencing_date
            )

    @handle_post_processing_errors(
        to_except=(PostProcessingDataTransferError, ValueError),
        to_raise=PostProcessingStoreDataError,
    )
    def store_post_processing_data(self, run_data: PacBioRunData, dry_run: bool = False) -> None:
        dtos: PacBioDTOs = self.data_transfer_service.get_post_processing_dtos(run_data)
        smrt_cell: PacbioSMRTCell = self._create_run_device(dtos.run_device)
        sequencing_run: PacbioSequencingRun = self._create_instrument_run(
            instrument_run_dto=dtos.sequencing_run, smrt_cell=smrt_cell
        )
        self._create_sample_run_metrics(
            sample_run_metrics_dtos=dtos.sample_sequencing_metrics, sequencing_run=sequencing_run
        )
        self._update_sample(
            sample_run_metrics_dtos=dtos.sample_sequencing_metrics,
            sequencing_date=sequencing_run.completed_at,
        )
        if dry_run:
            self.store.rollback()
            LOG.info(
                f"Dry run, no entries will be added to database for SMRT cell {run_data.full_path}."
            )
            return
        LOG.debug(f"Data stored in statusDB for run {run_data.sequencing_run_name}")
        self.store.commit_to_store()



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.rna_fusion.models.sample import RNAFusionSample

NewSample = Annotated[RNAFusionSample, Tag("new")]
OldSample = Annotated[ExistingSample, Tag("existing")]


class RNAFusionCase(Case):
    cohorts: list[str] | None = None
    synopsis: str | None = None
    samples: list[Annotated[NewSample | OldSample, Discriminator(has_internal_id)]]



from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.services.orders.order_summary_service.dto.order_summary import OrderSummary
from cg.services.orders.order_summary_service.utils import _get_analysis_map
from cg.store.models import Order
from cg.store.store import Store


class OrderSummaryService:
    def __init__(self, analysis_client: TrailblazerAPI, store: Store) -> None:
        self.analysis_client = analysis_client
        self.store = store

    def get_summary(self, order_id: int) -> OrderSummary:
        return self.get_summaries([order_id])[0]

    def get_summaries(self, order_ids: list[int]) -> list[OrderSummary]:
        orders: list[Order] = self.store.get_orders_by_ids(order_ids)
        analyses: list[AnalysisSummary] = self.analysis_client.get_summaries(order_ids)
        order_summaries = self.create_summaries(orders=orders, analysis_summaries=analyses)
        return order_summaries

    def create_summaries(
        self, orders: list[Order], analysis_summaries: list[AnalysisSummary]
    ) -> list[OrderSummary]:
        order_summaries: list[OrderSummary] = []
        analysis_summary_map: dict = _get_analysis_map(analysis_summaries)
        for order in orders:
            analysis_summary: AnalysisSummary = analysis_summary_map.get(order.id)
            order_summary = self.create_order_summary(order=order, summary=analysis_summary)
            order_summaries.append(order_summary)
        return order_summaries

    def create_order_summary(self, order: Order, summary: AnalysisSummary) -> OrderSummary:
        """Adds statuses inferred from StatusDB data for any cases not included in the provided summary."""
        order_id = order.id
        counted_cases: list[str] = summary.case_ids if summary else []
        not_received: int = self.store.get_case_not_received_count(
            order_id=order_id, cases_to_exclude=counted_cases
        )
        in_preparation: int = self.store.get_case_in_preparation_count(
            order_id=order_id, cases_to_exclude=counted_cases
        )
        in_sequencing: int = self.store.get_case_in_sequencing_count(
            order_id=order_id, cases_to_exclude=counted_cases
        )
        failed_sequencing_qc: int = self.store.get_case_failed_sequencing_count(
            order_id=order_id, cases_to_exclude=counted_cases
        )

        return OrderSummary(
            order_id=order_id,
            total=len(order.cases),
            cancelled=summary.cancelled.count,
            completed=summary.completed.count,
            delivered=summary.delivered.count,
            failed=summary.failed.count,
            failed_sequencing_qc=failed_sequencing_qc,
            in_lab_preparation=in_preparation,
            in_sequencing=in_sequencing,
            not_received=not_received,
            running=summary.running.count,
        )



from cg.services.orders.validation.models.order_with_samples import OrderWithSamples
from cg.services.orders.validation.order_types.microbial_fastq.constants import (
    MicrobialFastqDeliveryType,
)
from cg.services.orders.validation.order_types.microbial_fastq.models.sample import (
    MicrobialFastqSample,
)


class MicrobialFastqOrder(OrderWithSamples):
    delivery_type: MicrobialFastqDeliveryType
    samples: list[MicrobialFastqSample]

    @property
    def enumerated_samples(self) -> enumerate[MicrobialFastqSample]:
        return enumerate(self.samples)



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import Customer, User


def filter_user_by_email(users: Query, email: str, **kwargs) -> Query:
    """Return user by email."""
    return users.filter(User.email == email)


def filter_user_by_id(users: Query, user_id: int, **kwargs) -> Query:
    return users.filter(User.id == user_id)


def filter_user_by_customer_internal_id(users: Query, customer_internal_id: str, **kwargs) -> Query:
    return users.join(User.customers).filter(Customer.internal_id == customer_internal_id)


class UserFilter(Enum):
    """Define User filter functions."""

    BY_EMAIL: Callable = filter_user_by_email
    BY_ID: Callable = filter_user_by_id
    BY_CUSTOMER_INTERNAL_ID: Callable = filter_user_by_customer_internal_id


def apply_user_filter(
    users: Query,
    filter_functions: list[Callable],
    email: str | None = None,
    user_id: int | None = None,
    customer_internal_id: str | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for filter_function in filter_functions:
        users: Query = filter_function(
            users=users,
            email=email,
            user_id=user_id,
            customer_internal_id=customer_internal_id,
        )
    return users



from typing import Callable

from cg.models.orders.sample_base import ControlEnum
from cg.services.orders.validation.constants import ElutionBuffer, ExtractionMethod
from cg.services.orders.validation.errors.case_errors import CaseError
from cg.services.orders.validation.errors.case_sample_errors import CaseSampleError
from cg.services.orders.validation.errors.order_errors import OrderError
from cg.services.orders.validation.errors.sample_errors import SampleError
from cg.services.orders.validation.models.order import Order
from cg.store.store import Store


def apply_order_validation(rules: list[Callable], order: Order, store: Store) -> list[OrderError]:
    errors: list[OrderError] = []
    for rule in rules:
        rule_errors: list[OrderError] = rule(order=order, store=store)
        errors.extend(rule_errors)
    return errors


def apply_case_validation(rules: list[Callable], order: Order, store: Store) -> list[CaseError]:
    errors: list[CaseError] = []
    for rule in rules:
        rule_errors: list[CaseError] = rule(order=order, store=store)
        errors.extend(rule_errors)
    return errors


def apply_case_sample_validation(
    rules: list[Callable], order: Order, store: Store
) -> list[CaseSampleError]:
    errors: list[CaseSampleError] = []
    for rule in rules:
        rule_errors: list[CaseSampleError] = rule(order=order, store=store)
        errors.extend(rule_errors)
    return errors


def apply_sample_validation(rules: list[Callable], order: Order, store: Store) -> list[SampleError]:
    errors: list[SampleError] = []
    for rule in rules:
        rule_errors: list[SampleError] = rule(order=order, store=store)
        errors.extend(rule_errors)
    return errors


def parse_buffer(buffer: str | None) -> ElutionBuffer | None:
    return ElutionBuffer.OTHER if buffer and buffer.startswith("Other") else buffer


def parse_control(control: ControlEnum | None) -> ControlEnum:
    """Convert the control value into one of the Enum values if it's None."""
    return control or ControlEnum.not_control


def parse_extraction_method(extraction_method: str | None) -> ExtractionMethod:
    return (
        ExtractionMethod.MAGNAPURE_96
        if extraction_method and extraction_method.startswith(ExtractionMethod.MAGNAPURE_96)
        else extraction_method
    )



from wtforms.fields.choices import SelectMultipleField
from wtforms.widgets.core import CheckboxInput, TableWidget


class MultiCheckboxField(SelectMultipleField):
    widget = TableWidget()
    option_widget = CheckboxInput()



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.mip_dna.models.case import MIPDNACase
from cg.services.orders.validation.order_types.raredisease.models.sample import RarediseaseSample

NewSample = Annotated[RarediseaseSample, Tag("new")]
OldSample = Annotated[ExistingSample, Tag("existing")]


class RarediseaseCase(MIPDNACase):
    samples: list[Annotated[NewSample | OldSample, Discriminator(has_internal_id)]]

    def get_samples_with_father(self) -> list[tuple[RarediseaseSample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.father]

    def get_samples_with_mother(self) -> list[tuple[RarediseaseSample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.mother]



from pathlib import Path
from typing import Any, Type

from cg.constants.constants import FileFormat
from cg.constants.pacbio import MetricsFileFields, PacBioDirsAndFiles
from cg.io.controller import ReadFile
from cg.services.run_devices.pacbio.metrics_parser.models import (
    BarcodeMetrics,
    BaseMetrics,
    ControlMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    ReadMetrics,
    SampleMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.utils.files import get_file_with_pattern_from_list


def _get_data_model_from_pattern(pattern: str) -> Type[BaseMetrics]:
    """Return the data model based on the pattern."""
    pattern_to_model = {
        PacBioDirsAndFiles.BARCODES_REPORT: BarcodeMetrics,
        PacBioDirsAndFiles.CCS_REPORT_SUFFIX: ReadMetrics,
        PacBioDirsAndFiles.CONTROL_REPORT: ControlMetrics,
        PacBioDirsAndFiles.LOADING_REPORT: ProductivityMetrics,
        PacBioDirsAndFiles.RAW_DATA_REPORT: PolymeraseMetrics,
        PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT: SmrtlinkDatasetsMetrics,
    }
    return pattern_to_model.get(pattern)


def _parse_report_to_model(report_file: Path, data_model: Type[BaseMetrics]) -> BaseMetrics:
    """Parse the metrics report to a data model."""
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report_file)
    if data_model == SmrtlinkDatasetsMetrics:
        return data_model.model_validate(parsed_json[0])
    metrics: list[dict[str, Any]] = parsed_json.get(MetricsFileFields.ATTRIBUTES)
    data: dict = {
        report_field[MetricsFileFields.ID]: report_field[MetricsFileFields.VALUE]
        for report_field in metrics
    }
    return data_model.model_validate(data)


def get_parsed_metrics_from_file_name(metrics_files: list[Path], file_name: str) -> BaseMetrics:
    report_file: Path = get_file_with_pattern_from_list(files=metrics_files, pattern=file_name)
    data_model: Type[BaseMetrics] = _get_data_model_from_pattern(pattern=file_name)
    return _parse_report_to_model(report_file=report_file, data_model=data_model)


def _is_unassigned_reads_entry(sample_metric: SampleMetrics) -> bool:
    return sample_metric.sample_internal_id == "No Name"


def _parse_sample_data(sample_data: list[dict[str, Any]]) -> list[SampleMetrics]:
    """Parse all samples data into SampleMetrics given the sample section of the barcodes report."""
    sample_metrics: list[SampleMetrics] = []
    number_of_samples: int = len(sample_data[0].get(MetricsFileFields.VALUES))
    for sample_idx in range(number_of_samples):
        sample: dict = {}
        for data_field in sample_data:
            field_id: str = data_field.get(MetricsFileFields.ID)
            sample[field_id] = data_field.get(MetricsFileFields.VALUES)[sample_idx]
        sample_metric = SampleMetrics.model_validate(sample)
        if not _is_unassigned_reads_entry(sample_metric):
            sample_metrics.append(sample_metric)
    return sample_metrics


def get_parsed_sample_metrics(metrics_files: list[Path]) -> list[SampleMetrics]:
    """Parse the metrics from the barcodes report for each sample."""
    report_file: Path = get_file_with_pattern_from_list(
        files=metrics_files, pattern=PacBioDirsAndFiles.BARCODES_REPORT
    )
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report_file)
    sample_data: list[dict[str, Any]] = parsed_json.get(MetricsFileFields.TABLES)[0].get(
        MetricsFileFields.COLUMNS
    )
    return _parse_sample_data(sample_data)



import re
from collections import Counter

from cg.constants.constants import StatusOptions
from cg.constants.subject import Sex
from cg.models.orders.constants import OrderType
from cg.models.orders.sample_base import ContainerEnum, SexEnum
from cg.services.orders.validation.errors.case_errors import RepeatedCaseNameError
from cg.services.orders.validation.errors.case_sample_errors import (
    FatherNotInCaseError,
    InvalidConcentrationIfSkipRCError,
    InvalidFatherSexError,
    InvalidMotherSexError,
    MotherNotInCaseError,
    OccupiedWellError,
    SubjectIdSameAsCaseNameError,
)
from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.case_aliases import (
    CaseContainingRelatives,
    CaseWithSkipRC,
)
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.models.sample import Sample
from cg.services.orders.validation.models.sample_aliases import (
    HumanSample,
    SampleInCase,
    SampleWithRelatives,
)
from cg.services.orders.validation.rules.case.utils import is_sample_in_case
from cg.services.orders.validation.rules.utils import (
    get_concentration_interval,
    has_sample_invalid_concentration,
    is_in_container,
    is_sample_on_plate,
    is_volume_within_allowed_interval,
)
from cg.store.models import Customer
from cg.store.models import Sample as DbSample
from cg.store.store import Store


def is_concentration_missing(sample: SampleWithRelatives) -> bool:
    return not sample.concentration_ng_ul


def is_well_position_missing(sample: SampleWithRelatives) -> bool:
    return sample.container == ContainerEnum.plate and not sample.well_position


def is_container_name_missing(sample: SampleWithRelatives) -> bool:
    return sample.container == ContainerEnum.plate and not sample.container_name


def get_invalid_panels(panels: list[str], store: Store) -> list[str]:
    invalid_panels: list[str] = [
        panel for panel in panels if not store.does_gene_panel_exist(panel)
    ]
    return invalid_panels


def is_volume_invalid(sample: Sample) -> bool:
    in_container: bool = is_in_container(sample.container)
    allowed_volume: bool = is_volume_within_allowed_interval(sample.volume)
    return in_container and not allowed_volume


def get_well_sample_map(
    order: OrderWithCases, **kwargs
) -> dict[tuple[str, str], list[tuple[int, int]]]:
    """
    Constructs a dict with keys being a (container_name, well_position) pair. For each such pair, the value will be
    a list of (case index, sample index) pairs corresponding to all samples with matching container_name and
    well_position, provided the sample is on a plate.
    """
    well_position_to_sample_map = {}
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_sample_on_plate(sample):
                key: tuple[str, str] = (sample.container_name, sample.well_position)
                value: tuple[int, int] = (case_index, sample_index)
                if not well_position_to_sample_map.get(key):
                    well_position_to_sample_map[key] = []
                well_position_to_sample_map[key].append(value)
    return well_position_to_sample_map


def get_occupied_well_errors(colliding_samples: list[tuple[int, int]]) -> list[OccupiedWellError]:
    errors: list[OccupiedWellError] = []
    for case_index, sample_index in colliding_samples:
        error = OccupiedWellError(case_index=case_index, sample_index=sample_index)
        errors.append(error)
    return errors


def get_indices_for_repeated_case_names(order: OrderWithCases) -> list[int]:
    counter = Counter([case.name for _, case in order.enumerated_new_cases])
    indices: list[int] = []

    for index, case in order.enumerated_new_cases:
        if counter.get(case.name) > 1:
            indices.append(index)

    return indices


def get_repeated_case_name_errors(order: OrderWithCases) -> list[RepeatedCaseNameError]:
    case_indices: list[int] = get_indices_for_repeated_case_names(order)
    return [RepeatedCaseNameError(case_index=case_index) for case_index in case_indices]


def get_father_sex_errors(
    case: CaseContainingRelatives, case_index: int, store: Store
) -> list[InvalidFatherSexError]:
    errors: list[InvalidFatherSexError] = []
    children: list[tuple[SampleWithRelatives, int]] = case.get_samples_with_father()
    for child, child_index in children:
        if is_father_sex_invalid(child=child, case=case, store=store):
            error: InvalidFatherSexError = create_father_sex_error(
                case_index=case_index, sample_index=child_index
            )
            errors.append(error)
    return errors


def is_father_sex_invalid(
    child: SampleWithRelatives, case: CaseContainingRelatives, store: Store
) -> bool:
    father: SampleWithRelatives | None = case.get_new_sample(child.father)
    if not father:
        father: DbSample | None = case.get_existing_sample_from_db(
            sample_name=child.father, store=store
        )
    return father and father.sex != Sex.MALE


def create_father_sex_error(case_index: int, sample_index: int) -> InvalidFatherSexError:
    return InvalidFatherSexError(case_index=case_index, sample_index=sample_index)


def get_father_case_errors(
    case: CaseContainingRelatives, case_index: int, store: Store
) -> list[FatherNotInCaseError]:
    errors: list[FatherNotInCaseError] = []
    children: list[tuple[SampleWithRelatives | ExistingSample, int]] = (
        case.get_samples_with_father()
    )
    for child, child_index in children:
        if not is_sample_in_case(case=case, sample_name=child.father, store=store):
            error: FatherNotInCaseError = create_father_case_error(
                case_index=case_index,
                sample_index=child_index,
            )
            errors.append(error)
    return errors


def get_mother_sex_errors(
    case: CaseContainingRelatives, case_index: int, store: Store
) -> list[InvalidMotherSexError]:
    errors: list[InvalidMotherSexError] = []
    children: list[tuple[SampleWithRelatives, int]] = case.get_samples_with_mother()
    for child, child_index in children:
        if is_mother_sex_invalid(child=child, case=case, store=store):
            error: InvalidMotherSexError = create_mother_sex_error(
                case_index=case_index,
                sample_index=child_index,
            )
            errors.append(error)
    return errors


def get_mother_case_errors(
    case: CaseContainingRelatives, case_index: int, store: Store
) -> list[MotherNotInCaseError]:
    errors: list[MotherNotInCaseError] = []
    children: list[tuple[SampleWithRelatives, int]] = case.get_samples_with_mother()
    for child, child_index in children:
        if not is_sample_in_case(case=case, sample_name=child.mother, store=store):
            error: MotherNotInCaseError = create_mother_case_error(
                case_index=case_index, sample_index=child_index
            )
            errors.append(error)
    return errors


def create_father_case_error(case_index: int, sample_index: int) -> FatherNotInCaseError:
    return FatherNotInCaseError(case_index=case_index, sample_index=sample_index)


def create_mother_case_error(case_index: int, sample_index: int) -> MotherNotInCaseError:
    return MotherNotInCaseError(case_index=case_index, sample_index=sample_index)


def is_mother_sex_invalid(
    child: SampleWithRelatives, case: CaseContainingRelatives, store: Store
) -> bool:
    mother: SampleWithRelatives | None = case.get_new_sample(child.mother)
    if not mother:
        mother: DbSample | None = case.get_existing_sample_from_db(
            sample_name=child.mother, store=store
        )
    return mother and mother.sex != Sex.FEMALE


def create_mother_sex_error(case_index: int, sample_index: int) -> InvalidMotherSexError:
    return InvalidMotherSexError(case_index=case_index, sample_index=sample_index)


def has_sex_and_subject(sample: HumanSample) -> bool:
    return bool(sample.subject_id and sample.sex != SexEnum.unknown)


def validate_subject_ids_in_case(
    case: CaseContainingRelatives, case_index: int
) -> list[SubjectIdSameAsCaseNameError]:
    errors: list[SubjectIdSameAsCaseNameError] = []
    for sample_index, sample in case.enumerated_new_samples:
        if sample.subject_id == case.name:
            error = SubjectIdSameAsCaseNameError(case_index=case_index, sample_index=sample_index)
            errors.append(error)
    return errors


def validate_concentration_in_case(
    case: CaseWithSkipRC, case_index: int, store: Store
) -> list[InvalidConcentrationIfSkipRCError]:
    errors: list[InvalidConcentrationIfSkipRCError] = []
    for sample_index, sample in case.enumerated_new_samples:
        if application := store.get_application_by_tag(sample.application):
            allowed_interval = get_concentration_interval(sample=sample, application=application)
            if has_sample_invalid_concentration(sample=sample, allowed_interval=allowed_interval):
                error: InvalidConcentrationIfSkipRCError = create_invalid_concentration_error(
                    case_index=case_index,
                    sample_index=sample_index,
                    allowed_interval=allowed_interval,
                )
                errors.append(error)
    return errors


def create_invalid_concentration_error(
    case_index: int, sample_index: int, allowed_interval: tuple[float, float]
) -> InvalidConcentrationIfSkipRCError:
    return InvalidConcentrationIfSkipRCError(
        case_index=case_index,
        sample_index=sample_index,
        allowed_interval=allowed_interval,
    )


def is_invalid_plate_well_format(sample: Sample) -> bool:
    """Check if a sample has an invalid well format."""
    if sample.is_on_plate and sample.well_position:
        correct_well_position_pattern: str = r"^[A-H]:([1-9]|1[0-2])$"
        return not bool(re.match(correct_well_position_pattern, sample.well_position))
    return False


def is_sample_tube_name_reused(sample: Sample, counter: Counter) -> bool:
    """Check if a tube container name is reused across samples."""
    return sample.container == ContainerEnum.tube and counter.get(sample.container_name) > 1


def get_counter_container_names(order: OrderWithCases) -> Counter:
    counter = Counter(
        sample.container_name
        for case_index, case in order.enumerated_new_cases
        for sample_index, sample in case.enumerated_new_samples
    )
    return counter


def get_existing_sample_names(order: OrderWithCases, status_db: Store) -> set[str]:
    existing_sample_names: set[str] = set()
    for case in order.cases:
        if case.is_new:
            for sample_index, sample in case.enumerated_existing_samples:
                db_sample = status_db.get_sample_by_internal_id(sample.internal_id)
                existing_sample_names.add(db_sample.name)
        else:
            db_case = status_db.get_case_by_internal_id(case.internal_id)
            for sample in db_case.samples:
                existing_sample_names.add(sample.name)
    return existing_sample_names


def are_all_samples_unknown(case: Case) -> bool:
    """Check if all samples in a case are unknown."""
    return all(sample.status == StatusOptions.UNKNOWN for sample in case.samples)


def is_buffer_missing(sample: SampleInCase) -> bool:
    applications_requiring_buffer: tuple = ("PAN", "EX", "WGSWPF", "METWPF")
    return bool(
        sample.application.startswith(tuple(applications_requiring_buffer))
        and not sample.elution_buffer
    )


def is_sample_not_from_collaboration(
    customer_id: str, sample: ExistingSample, store: Store
) -> bool:
    db_sample: DbSample | None = store.get_sample_by_internal_id(sample.internal_id)
    customer: Customer | None = store.get_customer_by_internal_id(customer_id)
    return db_sample and customer and db_sample.customer not in customer.collaborators


def get_existing_case_names(order: OrderWithCases, status_db: Store) -> set[str]:
    existing_case_names: set[str] = set()
    for _, case in order.enumerated_existing_cases:
        if db_case := status_db.get_case_by_internal_id(case.internal_id):
            existing_case_names.add(db_case.name)
    return existing_case_names


def is_sample_compatible_with_order_type(
    order_type: OrderType, sample: ExistingSample, store: Store
) -> bool:
    if db_sample := store.get_sample_by_internal_id(sample.internal_id):
        return order_type in db_sample.application_version.application.order_types
    else:
        return True



"""Class to hold file information about a compression entity"""

import logging
import os
from datetime import date, datetime
from pathlib import Path

from cg.apps.crunchy.files import check_if_update_spring, get_crunchy_metadata, get_file_updated_at
from cg.apps.crunchy.models import CrunchyMetadata
from cg.constants import FASTQ_FIRST_READ_SUFFIX, FASTQ_SECOND_READ_SUFFIX, FileExtensions
from cg.constants.compression import PENDING_PATH_SUFFIX

LOG = logging.getLogger(__name__)


class CompressionData:
    """Holds information about compression data"""

    def __init__(self, stub: Path = None):
        """Initialise a compression data object

        The stub is first part of the file name
        """
        self.stub = stub
        self.stub_string = str(self.stub)

    @property
    def pending_path(self) -> Path:
        """Return the path to a compression pending file"""
        return self.stub.with_suffix(PENDING_PATH_SUFFIX)

    @property
    def spring_path(self) -> Path:
        """Return the path to a SPRING file"""
        return self.stub.with_suffix(FileExtensions.SPRING)

    @property
    def spring_metadata_path(self) -> Path:
        """Return the path to a SPRING metadata file"""
        return self.stub.with_suffix(".json")

    @property
    def analysis_dir(self) -> Path:
        """Return the path to folder where analysis is"""
        return self.stub.resolve().parent

    @property
    def fastq_first(self) -> Path:
        """Return the path to the first read in pair"""
        return Path(self.stub_string + FASTQ_FIRST_READ_SUFFIX)

    @property
    def fastq_second(self) -> Path:
        """Return the path to the second read in pair"""
        return Path(self.stub_string + FASTQ_SECOND_READ_SUFFIX)

    @property
    def run_name(self) -> str:
        """Return the name of the sequencing run identifier"""
        return self.stub.name

    def pair_exists(self) -> bool:
        """Check that both files in FASTQ pair exists"""
        LOG.info("Check if FASTQ pair exists")
        if not self.file_exists_and_is_accessible(self.fastq_first):
            return False
        return bool(self.file_exists_and_is_accessible(self.fastq_second))

    @staticmethod
    def is_absolute(file_path: Path) -> bool:
        """Check if file path can be resolved"""
        if not file_path.is_absolute():
            LOG.info("Could not resolve full path from HK to %s", file_path)
            return False
        return True

    @staticmethod
    def file_exists_and_is_accessible(file_path: Path) -> bool:
        """Check if file exists and is accesible"""
        try:
            if not file_path.exists():
                LOG.info("%s does not exist", file_path)
                return False
        except PermissionError:
            LOG.warning("Not permitted to access %s. Skipping", file_path)
            return False
        return True

    @staticmethod
    def is_symlink(file_path: Path) -> bool:
        """Check if file path is symbolik link"""
        LOG.info("Check if %s is a symlink", file_path)
        return os.path.islink(file_path)

    @staticmethod
    def get_change_date(file_path: Path) -> datetime:
        """Return the time when this file was changed"""
        changed_date = datetime.fromtimestamp(file_path.stat().st_mtime)
        LOG.info("File %s was changed %s", file_path, changed_date)
        return changed_date

    def spring_exists(self) -> bool:
        """Check if the SPRING file exists"""
        LOG.info("Check if SPRING archive file exists")
        return self.file_exists_and_is_accessible(self.spring_path)

    def metadata_exists(self) -> bool:
        """Check if the SPRING metadata file exists"""
        LOG.info("Check if SPRING metadata file exists")
        return self.file_exists_and_is_accessible(self.spring_metadata_path)

    def pending_exists(self) -> bool:
        """Check if the SPRING pending flag file exists"""
        LOG.info("Check if pending compression file exists")
        return self.file_exists_and_is_accessible(self.pending_path)

    @property
    def is_compression_pending(self) -> bool:
        """Check if compression/decompression has started but not finished."""
        if self.pending_exists():
            LOG.info(f"Compression/decompression is pending for {self.run_name}")
            return True
        LOG.info("Compression/decompression is not running")
        return False

    @property
    def is_fastq_compression_possible(self) -> bool:
        """Check if FASTQ compression is possible.

        - Compression is running          -> Compression NOT possible
        - SPRING file exists on Hasta     -> Compression NOT possible
        - Data is external                -> Compression NOT possible
        - Not compressed and not running  -> Compression IS possible
        """
        if self.is_compression_pending:
            return False

        if self.spring_exists():
            LOG.debug("SPRING file found")
            return False

        if "external-data" in str(self.fastq_first):
            LOG.debug("File is external data and should not be compressed")
            return False

        LOG.debug("FASTQ compression is possible")

        return True

    @property
    def is_spring_decompression_possible(self) -> bool:
        """Check if SPRING decompression is possible.

        There are three possible answers to this question:

            - Compression/Decompression is running      -> Decompression is NOT possible
            - The FASTQ files are not compressed        -> Decompression is NOT possible
            - Compression has been performed            -> Decompression IS possible

        """
        if self.pending_exists():
            LOG.info(f"Compression/decompression is pending for {self.run_name}")
            return False

        if not self.spring_exists():
            LOG.info("No SPRING file found")
            return False

        if self.pair_exists():
            LOG.info("FASTQ files already exists")
            return False

        LOG.info("Decompression is possible")

        return True

    @property
    def is_fastq_compression_done(self) -> bool:
        """Check if FASTQ compression is finished.

        This is checked by controlling that the SPRING files that are produced after FASTQ
        compression exists.

        The following has to be fulfilled for FASTQ compression to be considered done:

            - A SPRING archive file exists
            - A SPRING archive metadata file exists
            - The SPRING archive has not been unpacked before FASTQ delta (21 days)

        Note:
        'updated_at' indicates at what date the SPRING archive was unarchived last.
        If the SPRING archive has never been unarchived 'updated_at' is None.

        """
        LOG.info("Check if FASTQ compression is finished")
        LOG.info(f"Check if SPRING file {self.spring_path} exists")
        if not self.spring_exists():
            LOG.info(
                f"No SPRING file for {self.run_name}",
            )
            return False
        LOG.info("SPRING file found")

        LOG.info(f"Check if SPRING metadata file {self.spring_metadata_path} exists")
        if not self.metadata_exists():
            LOG.info("No metadata file found")
            return False
        LOG.info("SPRING metadata file found")

        # We want this to raise exception if file is malformed
        crunchy_metadata: CrunchyMetadata = get_crunchy_metadata(self.spring_metadata_path)

        # Check if the SPRING archive has been unarchived
        updated_at: date | None = get_file_updated_at(crunchy_metadata)
        if updated_at is None:
            LOG.info(f"FASTQ compression is done for {self.run_name}")
            return True

        LOG.info(f"Files where unpacked {updated_at}")

        if not check_if_update_spring(updated_at):
            return False

        LOG.info(f"FASTQ compression is done for {self.run_name}")

        return True

    @property
    def is_spring_decompression_done(self) -> bool:
        """Check if SPRING decompression if finished.

        This means that all three files specified in SPRING metadata should exist.
        That is

            - First read in FASTQ pair should exist
            - Second read in FASTQ pair should exist
            - SPRING archive file should still exist
        """

        spring_metadata_path: Path = self.spring_metadata_path
        LOG.info(f"Check if SPRING metadata file {spring_metadata_path} exists")

        if not self.metadata_exists():
            LOG.info("No SPRING metadata file found")
            return False

        # We want this to exit hard if the metadata is malformed
        crunchy_metadata: CrunchyMetadata = get_crunchy_metadata(spring_metadata_path)

        for file_info in crunchy_metadata.files:
            if not Path(file_info.path).exists():
                LOG.info(f"File {file_info.path} does not exist")
                return False
            if not file_info.updated:
                LOG.info("Files have not been unarchived")
                return False

        LOG.info(f"SPRING decompression is done for run {self.run_name}")
        return True

    def __str__(self):
        return f"CompressionData(run:{self.run_name})"

    def __repr__(self):
        return f"CompressionData(stub:{self.stub})"


class SampleCompressionData:
    """Object encapsulating a sample's compression status."""

    def __init__(self, sample_id: str, compression_objects: list[CompressionData]):
        self.sample_id = sample_id
        self.compression_objects = compression_objects

    def is_decompression_needed(self) -> bool:
        """Check if decompression is needed for the specified sample."""
        LOG.debug(f"Checking if decompression is needed for {self.sample_id}.")
        return any(
            not compression_object.is_compression_pending and not compression_object.pair_exists()
            for compression_object in self.compression_objects
        )

    def is_spring_decompression_running(self) -> bool:
        """Check if sample is being decompressed"""
        return any(
            compression_object.is_compression_pending
            for compression_object in self.compression_objects
        )

    def can_be_decompressed(self) -> bool:
        """Returns True if at least one Spring file can be decompressed, otherwise False"""
        return any(
            compression_object.is_spring_decompression_possible
            for compression_object in self.compression_objects
        )


class CaseCompressionData:
    """Object encapsulating a case's compression status."""

    def __init__(self, case_id: str, sample_compression_data: list[SampleCompressionData]):
        self.case_id = case_id
        self.sample_compression_data = sample_compression_data

    def is_spring_decompression_needed(self) -> bool:
        """Check if spring decompression needs to be started"""
        return any(
            sample_compression.is_decompression_needed()
            for sample_compression in self.sample_compression_data
        )

    def is_spring_decompression_running(self) -> bool:
        """Check if case is being decompressed"""
        return any(
            sample_compression.is_spring_decompression_running()
            for sample_compression in self.sample_compression_data
        )

    def can_at_least_one_sample_be_decompressed(self) -> bool:
        """Returns True if at least one sample can be decompressed, otherwise False"""
        return any(
            sample_compression.can_be_decompressed()
            for sample_compression in self.sample_compression_data
        )



from pathlib import Path

from cg.constants import FileExtensions
from cg.exc import ValidationError


def validated_pdc_encryption_key_path(value: Path) -> Path:
    if not value.name.endswith(f"{FileExtensions.KEY}{FileExtensions.GPG}"):
        raise ValidationError("Missing a valid encryption key.")
    return value


def validated_pdc_sequencing_file_path(value: Path) -> Path:
    if not value.name.endswith(f"{FileExtensions.TAR}{FileExtensions.GZIP}{FileExtensions.GPG}"):
        raise ValidationError("Missing a valid sequence file.")
    return value



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_sample import ExistingSample
from cg.services.orders.validation.order_types.mip_dna.models.sample import MIPDNASample

NewSample = Annotated[MIPDNASample, Tag("new")]
OldSample = Annotated[ExistingSample, Tag("existing")]


class MIPDNACase(Case):
    cohorts: list[str] | None = None
    panels: list[str]
    synopsis: str | None = None
    samples: list[Annotated[NewSample | OldSample, Discriminator(has_internal_id)]]

    def get_samples_with_father(self) -> list[tuple[MIPDNASample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.father]

    def get_samples_with_mother(self) -> list[tuple[MIPDNASample | ExistingSample, int]]:
        return [(sample, index) for index, sample in self.enumerated_samples if sample.mother]



from pydantic import BaseModel


class IlluminaSequencingMetricsDTO(BaseModel):
    flow_cell_name: str
    flow_cell_lane_number: int
    sample_internal_id: str
    sample_total_reads_in_lane: int
    sample_base_percentage_passing_q30: float


class PacbioSequencingMetricsDTO(BaseModel):
    hifi_mean_read_length: int
    hifi_median_read_quality: str
    hifi_reads: int
    hifi_yield: int
    sample_id: str
    smrt_cell_id: str



"""Handler to create data objects."""

import logging
from datetime import datetime

import petname
from sqlalchemy import Insert
from sqlalchemy.orm import Session

from cg.constants import DataDelivery, Priority, Workflow
from cg.constants.archiving import PDC_ARCHIVE_LOCATION
from cg.models.orders.constants import OrderType
from cg.services.illumina.data_transfer.models import (
    IlluminaFlowCellDTO,
    IlluminaSampleSequencingMetricsDTO,
    IlluminaSequencingRunDTO,
)
from cg.services.run_devices.pacbio.data_transfer_service.dto import (
    PacBioSampleSequencingMetricsDTO,
    PacBioSequencingRunDTO,
    PacBioSMRTCellDTO,
)
from cg.store.crud.read import ReadHandler
from cg.store.database import get_session
from cg.store.exc import EntryAlreadyExistsError, EntryNotFoundError
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    ApplicationVersion,
    Bed,
    BedVersion,
    Case,
    CaseSample,
    Collaboration,
    Customer,
    IlluminaFlowCell,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Invoice,
    Order,
    OrderTypeApplication,
    Organism,
    PacbioSampleSequencingMetrics,
    PacbioSequencingRun,
    PacbioSMRTCell,
    Panel,
    Pool,
    Sample,
    User,
    order_case,
)

LOG = logging.getLogger(__name__)


class CreateMixin(ReadHandler):
    """Methods related to adding new data to the store."""

    def generate_readable_sample_id(self) -> str:
        """Generates a petname as sample internal id for development purposes. Not used in normal production flow."""
        while True:
            random_id: str = petname.Generate(3, separator="")
            if not self.get_sample_by_internal_id(random_id):
                return random_id

    def generate_readable_case_id(self) -> str:
        while True:
            random_id: str = petname.Generate(2, separator="", letters=10)
            if not self.get_case_by_internal_id(random_id):
                return random_id

    def add_customer(
        self,
        internal_id: str,
        name: str,
        invoice_address: str,
        invoice_reference: str,
        data_archive_location: str = PDC_ARCHIVE_LOCATION,
        scout_access: bool = False,
        is_clinical: bool = False,
        *args,
        **kwargs,
    ) -> Customer:
        """Build a new customer record."""

        return Customer(
            internal_id=internal_id,
            name=name,
            scout_access=scout_access,
            invoice_address=invoice_address,
            invoice_reference=invoice_reference,
            data_archive_location=data_archive_location,
            is_clinical=is_clinical,
            **kwargs,
        )

    def add_collaboration(self, internal_id: str, name: str, **kwargs) -> Collaboration:
        """Build a new customer group record."""

        return Collaboration(internal_id=internal_id, name=name, **kwargs)

    def add_user(self, customer: Customer, email: str, name: str, is_admin: bool = False) -> User:
        """Build a new user record."""

        new_user = User(name=name, email=email, is_admin=is_admin)
        new_user.customers.append(customer)
        return new_user

    def add_application(
        self,
        tag: str,
        prep_category: str,
        description: str,
        percent_kth: int,
        percent_reads_guaranteed: int,
        is_accredited: bool = False,
        min_sequencing_depth: int = 0,
        **kwargs,
    ) -> Application:
        """Build a new application record."""

        return Application(
            tag=tag,
            prep_category=prep_category,
            description=description,
            is_accredited=is_accredited,
            min_sequencing_depth=min_sequencing_depth,
            percent_kth=percent_kth,
            percent_reads_guaranteed=percent_reads_guaranteed,
            **kwargs,
        )

    def link_order_types_to_application(
        self, application: Application, order_types: list[OrderType]
    ) -> list[OrderTypeApplication]:
        new_orders: list = []
        for order_type in order_types:
            new_record = OrderTypeApplication(application=application, order_type=order_type)
            new_orders.append(new_record)
        return new_orders

    def add_application_version(
        self,
        application: Application,
        version: int,
        valid_from: datetime,
        prices: dict,
        **kwargs,
    ) -> ApplicationVersion:
        """Build a new application version record."""

        new_record = ApplicationVersion(version=version, valid_from=valid_from, **kwargs)
        for price_key in [
            Priority.standard.name,
            Priority.priority.name,
            Priority.express.name,
            Priority.research.name,
        ]:
            setattr(new_record, f"price_{price_key}", prices[price_key])
        new_record.application = application
        return new_record

    @staticmethod
    def add_application_limitation(
        application: Application,
        workflow: str,
        limitations: str,
        comment: str = "Dummy comment",
        created_at: datetime = datetime.now(),
        updated_at: datetime = datetime.now(),
        **kwargs,
    ) -> ApplicationLimitations:
        """Build a new application limitations record."""
        return ApplicationLimitations(
            application=application,
            workflow=workflow,
            limitations=limitations,
            comment=comment,
            created_at=created_at,
            updated_at=updated_at,
            **kwargs,
        )

    def add_bed(self, name: str) -> Bed:
        """Build a new bed record."""
        return Bed(name=name)

    def add_bed_version(self, bed: Bed, version: int, filename: str, shortname: str) -> BedVersion:
        """Build a new bed version record."""
        bed_version: BedVersion = BedVersion(
            version=version, filename=filename, shortname=shortname
        )
        bed_version.bed = bed
        return bed_version

    def add_sample(
        self,
        name: str,
        sex: str,
        comment: str = None,
        control: str = None,
        downsampled_to: int = None,
        internal_id: str = None,
        last_sequenced_at: datetime = None,
        order: str = None,
        ordered: datetime = None,
        prepared_at: datetime = None,
        priority: Priority = None,
        received: datetime = None,
        original_ticket: str = None,
        tumour: bool = False,
        subject_id: str = None,
        **kwargs,
    ) -> Sample:
        """Build a new Sample record."""

        internal_id = internal_id or self.generate_readable_sample_id()
        priority = priority or (Priority.research if downsampled_to else Priority.standard)
        return Sample(
            comment=comment,
            control=control,
            downsampled_to=downsampled_to,
            internal_id=internal_id,
            is_tumour=tumour,
            last_sequenced_at=last_sequenced_at,
            name=name,
            order=order,
            ordered_at=ordered or datetime.now(),
            original_ticket=original_ticket,
            prepared_at=prepared_at,
            priority=priority,
            received_at=received,
            sex=sex,
            subject_id=subject_id,
            **kwargs,
        )

    def add_case(
        self,
        data_analysis: Workflow,
        data_delivery: DataDelivery,
        name: str,
        ticket: str,
        panels: list[str] | None = None,
        cohorts: list[str] | None = None,
        priority: Priority | None = Priority.standard,
        synopsis: str | None = None,
        customer_id: int | None = None,
        comment: str | None = None,
    ) -> Case:
        """Build a new Case record."""

        internal_id: str = self.generate_readable_case_id()
        return Case(
            comment=comment,
            cohorts=cohorts,
            data_analysis=str(data_analysis),
            data_delivery=str(data_delivery),
            internal_id=internal_id,
            name=name,
            panels=panels,
            priority=priority,
            synopsis=synopsis,
            tickets=ticket,
            customer_id=customer_id,
        )

    def relate_sample(
        self,
        case: Case,
        sample: Sample,
        status: str,
        mother: Sample = None,
        father: Sample = None,
    ) -> CaseSample:
        """Relate a sample record to a family record."""

        new_record: CaseSample = CaseSample(status=status)
        new_record.case = case
        new_record.sample = sample
        new_record.mother = mother
        new_record.father = father
        return new_record

    def add_analysis(
        self,
        workflow: Workflow,
        version: str | None = None,
        completed_at: datetime | None = None,
        primary: bool = False,
        uploaded: datetime | None = None,
        started_at: datetime | None = None,
        comment: str | None = None,
        **kwargs,
    ) -> Analysis:
        """Build a new Analysis record."""
        return Analysis(
            workflow=workflow,
            workflow_version=version,
            completed_at=completed_at,
            is_primary=primary,
            uploaded_at=uploaded,
            started_at=started_at,
            comment=comment,
            **kwargs,
        )

    def add_panel(
        self,
        customer: Customer,
        name: str,
        abbrev: str,
        version: float,
        date: datetime = None,
        genes: int = None,
    ) -> Panel:
        """Build a new panel record."""

        new_record = Panel(
            name=name, abbrev=abbrev, current_version=version, date=date, gene_count=genes
        )
        new_record.customer = customer
        return new_record

    def add_pool(
        self,
        customer: Customer,
        name: str,
        order: str,
        ordered: datetime,
        application_version: ApplicationVersion,
        ticket: str = None,
        comment: str = None,
        received_at: datetime = None,
        invoice_id: int = None,
        no_invoice: bool = None,
        delivered_at: datetime = None,
    ) -> Pool:
        """Build a new Pool record."""

        new_record: Pool = Pool(
            name=name,
            ordered_at=ordered or datetime.now(),
            order=order,
            ticket=ticket,
            received_at=received_at,
            comment=comment,
            delivered_at=delivered_at,
            invoice_id=invoice_id,
            no_invoice=no_invoice,
        )
        new_record.customer = customer
        new_record.application_version = application_version
        return new_record

    def add_invoice(
        self,
        customer: Customer,
        samples: list[Sample] = None,
        microbial_samples: list[Sample] = None,
        pools: list[Pool] = None,
        comment: str = None,
        discount: int = 0,
        record_type: str = None,
        invoiced_at: datetime | None = None,
    ):
        """Build a new Invoice record."""

        new_id = self.new_invoice_id()
        new_invoice: Invoice = Invoice(
            comment=comment,
            discount=discount,
            id=new_id,
            record_type=record_type,
            invoiced_at=invoiced_at,
        )
        new_invoice.customer = customer
        for sample in samples or []:
            new_invoice.samples.append(sample)
        for microbial_sample in microbial_samples or []:
            new_invoice.samples.append(microbial_sample)
        for pool in pools or []:
            new_invoice.pools.append(pool)
        return new_invoice

    def add_organism(
        self,
        internal_id: str,
        name: str,
        reference_genome: str = None,
        verified: bool = False,
        **kwargs,
    ) -> Organism:
        """Build a new Organism record."""
        return Organism(
            internal_id=internal_id,
            name=name,
            reference_genome=reference_genome,
            verified=verified,
            **kwargs,
        )

    def add_order(self, customer: Customer, ticket_id: int, **kwargs) -> Order:
        """Build a new Order record."""
        order = Order(
            customer=customer,
            order_date=datetime.now(),
            ticket_id=ticket_id,
            **kwargs,
        )
        return order

    @staticmethod
    def link_case_to_order(order_id: int, case_id: int):
        insert_statement: Insert = order_case.insert().values(order_id=order_id, case_id=case_id)
        session: Session = get_session()
        session.execute(insert_statement)
        session.commit()

    def add_illumina_flow_cell(self, flow_cell_dto: IlluminaFlowCellDTO) -> IlluminaFlowCell:
        """Add a new Illumina flow cell to the status database as a pending transaction."""
        try:
            self.get_illumina_flow_cell_by_internal_id(flow_cell_dto.internal_id)
        except EntryNotFoundError:
            new_flow_cell = IlluminaFlowCell(
                internal_id=flow_cell_dto.internal_id,
                type=flow_cell_dto.type,
                model=flow_cell_dto.model,
            )
            self.add_item_to_store(new_flow_cell)
            LOG.debug(f"Flow cell added to status db: {new_flow_cell.internal_id}.")
            return new_flow_cell
        raise EntryAlreadyExistsError(f"Flow cell already exists: {flow_cell_dto.internal_id}")

    def add_illumina_sequencing_run(
        self, sequencing_run_dto: IlluminaSequencingRunDTO, flow_cell: IlluminaFlowCell
    ) -> IlluminaSequencingRun:
        """Add a new Illumina flow cell to the status database as a pending transaction."""
        new_sequencing_run = IlluminaSequencingRun(
            type=sequencing_run_dto.type,
            device=flow_cell,
            sequencer_type=sequencing_run_dto.sequencer_type,
            sequencer_name=sequencing_run_dto.sequencer_name,
            data_availability=sequencing_run_dto.data_availability,
            archived_at=sequencing_run_dto.archived_at,
            has_backup=sequencing_run_dto.has_backup,
            total_reads=sequencing_run_dto.total_reads,
            total_undetermined_reads=sequencing_run_dto.total_undetermined_reads,
            percent_undetermined_reads=sequencing_run_dto.percent_undetermined_reads,
            percent_q30=sequencing_run_dto.percent_q30,
            mean_quality_score=sequencing_run_dto.mean_quality_score,
            total_yield=sequencing_run_dto.total_yield,
            yield_q30=sequencing_run_dto.yield_q30,
            cycles=sequencing_run_dto.cycles,
            demultiplexing_software=sequencing_run_dto.demultiplexing_software,
            demultiplexing_software_version=sequencing_run_dto.demultiplexing_software_version,
            sequencing_started_at=sequencing_run_dto.sequencing_started_at,
            sequencing_completed_at=sequencing_run_dto.sequencing_completed_at,
            demultiplexing_started_at=sequencing_run_dto.demultiplexing_started_at,
            demultiplexing_completed_at=sequencing_run_dto.demultiplexing_completed_at,
        )
        self.add_item_to_store(new_sequencing_run)
        LOG.debug(f"Sequencing run added to status db: {new_sequencing_run.device.internal_id}.")
        return new_sequencing_run

    def add_illumina_sample_metrics_entry(
        self, metrics_dto: IlluminaSampleSequencingMetricsDTO, sequencing_run: IlluminaSequencingRun
    ) -> IlluminaSampleSequencingMetrics:
        """
        Add a new Illumina Sample Sequencing Metrics entry to the status database as a pending
        transaction.
        """
        sample: Sample = self.get_sample_by_internal_id(metrics_dto.sample_id)
        if not sample:
            self.rollback()
            raise EntryNotFoundError(f"Sample not found: {metrics_dto.sample_id}")
        new_metric = IlluminaSampleSequencingMetrics(
            sample=sample,
            instrument_run=sequencing_run,
            type=metrics_dto.type,
            flow_cell_lane=metrics_dto.flow_cell_lane,
            total_reads_in_lane=metrics_dto.total_reads_in_lane,
            base_passing_q30_percent=metrics_dto.base_passing_q30_percent,
            base_mean_quality_score=metrics_dto.base_mean_quality_score,
            yield_=metrics_dto.yield_,
            yield_q30=metrics_dto.yield_q30,
            created_at=metrics_dto.created_at,
        )
        self.add_item_to_store(new_metric)
        return new_metric

    def create_pac_bio_smrt_cell(self, run_device_dto: PacBioSMRTCellDTO) -> PacbioSMRTCell:
        LOG.debug(f"Creating Pacbio SMRT cell for {run_device_dto.internal_id}")
        if self.get_pac_bio_smrt_cell_by_internal_id(run_device_dto.internal_id):
            raise ValueError(f"SMRT cell with {run_device_dto.internal_id} already exists.")
        new_smrt_cell = PacbioSMRTCell(
            type=run_device_dto.type, internal_id=run_device_dto.internal_id
        )
        self.add_item_to_store(new_smrt_cell)
        return new_smrt_cell

    def create_pac_bio_sequencing_run(
        self, sequencing_run_dto: PacBioSequencingRunDTO, smrt_cell: PacbioSMRTCell
    ) -> PacbioSequencingRun:
        LOG.debug(f"Creating Pacbio sequencing run for SMRT cell {smrt_cell.internal_id}")
        new_sequencing_run = PacbioSequencingRun(
            type=sequencing_run_dto.type,
            well=sequencing_run_dto.well,
            plate=sequencing_run_dto.plate,
            run_name=sequencing_run_dto.run_name,
            movie_name=sequencing_run_dto.movie_name,
            started_at=sequencing_run_dto.started_at,
            completed_at=sequencing_run_dto.completed_at,
            hifi_reads=sequencing_run_dto.hifi_reads,
            hifi_yield=sequencing_run_dto.hifi_yield,
            hifi_mean_read_length=sequencing_run_dto.hifi_mean_read_length,
            hifi_median_read_quality=sequencing_run_dto.hifi_median_read_quality,
            percent_reads_passing_q30=sequencing_run_dto.percent_reads_passing_q30,
            productive_zmws=sequencing_run_dto.productive_zmws,
            p0_percent=sequencing_run_dto.p0_percent,
            p1_percent=sequencing_run_dto.p1_percent,
            p2_percent=sequencing_run_dto.p2_percent,
            polymerase_mean_read_length=sequencing_run_dto.polymerase_mean_read_length,
            polymerase_read_length_n50=sequencing_run_dto.polymerase_read_length_n50,
            polymerase_mean_longest_subread=sequencing_run_dto.polymerase_mean_longest_subread,
            polymerase_longest_subread_n50=sequencing_run_dto.polymerase_longest_subread_n50,
            control_reads=sequencing_run_dto.control_reads,
            control_mean_read_length=sequencing_run_dto.control_mean_read_length,
            control_mean_read_concordance=sequencing_run_dto.control_mean_read_concordance,
            control_mode_read_concordance=sequencing_run_dto.control_mode_read_concordance,
            failed_reads=sequencing_run_dto.failed_reads,
            failed_yield=sequencing_run_dto.failed_yield,
            failed_mean_read_length=sequencing_run_dto.failed_mean_read_length,
            barcoded_hifi_reads=sequencing_run_dto.barcoded_hifi_reads,
            barcoded_hifi_reads_percentage=sequencing_run_dto.barcoded_hifi_reads_percentage,
            barcoded_hifi_yield=sequencing_run_dto.barcoded_hifi_yield,
            barcoded_hifi_yield_percentage=sequencing_run_dto.barcoded_hifi_yield_percentage,
            barcoded_hifi_mean_read_length=sequencing_run_dto.barcoded_hifi_mean_read_length,
            unbarcoded_hifi_reads=sequencing_run_dto.unbarcoded_hifi_reads,
            unbarcoded_hifi_yield=sequencing_run_dto.unbarcoded_hifi_yield,
            unbarcoded_hifi_mean_read_length=sequencing_run_dto.unbarcoded_hifi_mean_read_length,
            device=smrt_cell,
        )
        self.add_item_to_store(new_sequencing_run)
        return new_sequencing_run

    def create_pac_bio_sample_sequencing_run(
        self,
        sample_run_metrics_dto: PacBioSampleSequencingMetricsDTO,
        sequencing_run: PacbioSequencingRun,
    ) -> PacbioSampleSequencingMetrics:
        sample_id: str = sample_run_metrics_dto.sample_internal_id
        LOG.debug(f"Creating Pacbio sample sequencing metric for sample {sample_id}")
        sample: Sample = self.get_sample_by_internal_id(sample_id)
        if not sample:
            self.rollback()
            raise EntryNotFoundError(f"Sample not found: {sample_id}")
        new_sample_sequencing_run = PacbioSampleSequencingMetrics(
            sample=sample,
            hifi_reads=sample_run_metrics_dto.hifi_reads,
            hifi_yield=sample_run_metrics_dto.hifi_yield,
            hifi_mean_read_length=sample_run_metrics_dto.hifi_mean_read_length,
            hifi_median_read_quality=sample_run_metrics_dto.hifi_median_read_quality,
            instrument_run=sequencing_run,
            polymerase_mean_read_length=sample_run_metrics_dto.polymerase_mean_read_length,
        )
        self.add_item_to_store(new_sample_sequencing_run)
        return new_sample_sequencing_run



"""Commands to start MIP rare disease DNA workflow."""

import logging
from typing import cast

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.commands import (
    ensure_illumina_runs_on_disk,
    link,
    resolve_compression,
    store,
    store_available,
)
from cg.cli.workflow.mip.options import (
    ARGUMENT_CASE_ID,
    OPTION_BWA_MEM,
    OPTION_PANEL_BED,
    START_AFTER_PROGRAM,
    START_WITH_PROGRAM,
)
from cg.constants import Workflow
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.configurator.implementations.mip_dna import MIPDNAConfigurator
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory
from cg.services.analysis_starter.service import AnalysisStarter

LOG = logging.getLogger(__name__)


@click.group(
    "mip-dna",
    invoke_without_command=True,
    context_settings=CLICK_CONTEXT_SETTINGS,
)
@click.pass_context
def mip_dna(
    context: click.Context,
):
    """Rare disease DNA workflow"""
    AnalysisAPI.get_help(context)

    context.obj.meta_apis["analysis_api"] = MipDNAAnalysisAPI(config=context.obj)


for sub_cmd in [
    ensure_illumina_runs_on_disk,
    link,
    resolve_compression,
    store,
    store_available,
]:
    mip_dna.add_command(sub_cmd)


@mip_dna.command("run")
@START_AFTER_PROGRAM
@START_WITH_PROGRAM
@OPTION_BWA_MEM
@ARGUMENT_CASE_ID
@click.pass_obj
def run(
    cg_config: CGConfig,
    case_id: str,
    use_bwa_mem: bool,
    start_after: str | None,
    start_with: str | None,
):
    """
    Run a preconfigured MIP-DNA case.

    \b
    Assumes that the following files exist in the case run directory:
        - pedigree.yaml
        - gene_panels.bed
        - managed_variants.vcf
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(Workflow.MIP_DNA)
    analysis_starter.run(
        case_id=case_id, start_after=start_after, start_with=start_with, use_bwa_mem=use_bwa_mem
    )


@mip_dna.command("start")
@START_AFTER_PROGRAM
@START_WITH_PROGRAM
@OPTION_BWA_MEM
@OPTION_PANEL_BED
@ARGUMENT_CASE_ID
@click.pass_obj
def start(
    cg_config: CGConfig,
    case_id: str,
    use_bwa_mem: bool,
    panel_bed: str | None,
    start_after: str | None,
    start_with: str | None,
):
    """
    Start a MIP-DNA case.

    \b
    Configures the case and writes the following files:
        - pedigree.yaml
        - gene_panels.bed
        - managed_variants.vcf
    """
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(Workflow.MIP_DNA)
    analysis_starter.start(
        case_id=case_id,
        start_after=start_after,
        start_with=start_with,
        panel_bed=panel_bed,
        use_bwa_mem=use_bwa_mem,
    )


@mip_dna.command("start-available")
@click.pass_obj
def start_available(cg_config: CGConfig):
    """
    Starts all available MIP-DNA cases.

    \b
    Configures the individual case and writes the following files for each case:
        - pedigree.yaml
        - gene_panels.bed
        - managed_variants.vcf
    """
    LOG.info("Starting MIP-DNA workflow for all available cases.")
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(Workflow.MIP_DNA)
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort


@mip_dna.command("config-case")
@OPTION_PANEL_BED
@ARGUMENT_CASE_ID
@click.pass_obj
def config_case(
    cg_config: CGConfig,
    case_id: str,
    panel_bed: str | None,
):
    """
    Configure a MIP-DNA case so that it is ready to be run.

    \b
    This creates the following files:
        - pedigree.yaml
        - gene_panels.bed
        - managed_variants.vcf
    """
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(MIPDNAConfigurator, factory.get_configurator(Workflow.MIP_DNA))
    configurator.configure(case_id=case_id, panel_bed=panel_bed)



from pydantic import AfterValidator, BeforeValidator, Field
from typing_extensions import Annotated

from cg.models.orders.constants import ExcelSampleAliases
from cg.models.orders.sample_base import OrderSample
from cg.models.orders.validators.excel_sample_validators import (
    convert_sex,
    convert_to_date,
    convert_to_lower,
    convert_to_priority,
    empty_string_to_none,
    numeric_value,
    parse_panels,
    replace_spaces_with_underscores,
    validate_data_analysis,
    validate_parent,
    validate_source,
)


class ExcelSample(OrderSample):
    age_at_sampling: str = Field(None, alias=ExcelSampleAliases.AGE_AT_SAMPLING)
    application: str = Field(..., alias=ExcelSampleAliases.APPLICATION)
    capture_kit: str = Field(None, alias=ExcelSampleAliases.CAPTURE_KIT)
    cohorts: list[str] = Field(None, alias=ExcelSampleAliases.COHORTS)
    collection_date: Annotated[str, AfterValidator(convert_to_date)] = Field(
        None, alias=ExcelSampleAliases.COLLECTION_DATE
    )
    comment: str = Field(None, alias=ExcelSampleAliases.COMMENT)
    concentration: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias=ExcelSampleAliases.CONCENTRATION
    )
    concentration_ng_ul: str = Field(None, alias=ExcelSampleAliases.CONCENTRATION_NG_UL)
    concentration_sample: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias=ExcelSampleAliases.CONCENTRATION_SAMPLE
    )
    container: str = Field(None, alias=ExcelSampleAliases.CONTAINER)
    container_name: str = Field(None, alias=ExcelSampleAliases.CONTAINER_NAME)
    control: str = Field(None, alias=ExcelSampleAliases.CONTROL)
    custom_index: str = Field(None, alias=ExcelSampleAliases.CUSTOM_INDEX)
    customer: str = Field(..., alias=ExcelSampleAliases.CUSTOMER)
    data_analysis: Annotated[str, AfterValidator(validate_data_analysis)] = Field(
        "MIP DNA", alias=ExcelSampleAliases.DATA_ANALYSIS
    )
    data_delivery: Annotated[str, AfterValidator(convert_to_lower)] = Field(
        None, alias=ExcelSampleAliases.DATA_DELIVERY
    )
    elution_buffer: str = Field(None, alias=ExcelSampleAliases.ELUTION_BUFFER)
    extraction_method: str = Field(None, alias=ExcelSampleAliases.EXTRACTION_METHOD)
    family_name: str = Field(None, alias=ExcelSampleAliases.FAMILY_NAME)
    father: Annotated[str, AfterValidator(validate_parent)] = Field(
        None, alias=ExcelSampleAliases.FATHER
    )
    formalin_fixation_time: str = Field(None, alias=ExcelSampleAliases.FORMALIN_FIXATION_TIME)
    index: str = Field(None, alias=ExcelSampleAliases.INDEX)
    index_number: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias=ExcelSampleAliases.INDEX_NUMBER
    )
    lab_code: str = Field(None, alias=ExcelSampleAliases.LAB_CODE)
    mother: Annotated[str, AfterValidator(validate_parent)] = Field(
        None, alias=ExcelSampleAliases.MOTHER
    )
    name: str = Field(..., alias=ExcelSampleAliases.NAME)
    organism: str = Field(None, alias=ExcelSampleAliases.ORGANISM)
    organism_other: str = Field(None, alias=ExcelSampleAliases.ORGANISM_OTHER)
    original_lab: str = Field(None, alias=ExcelSampleAliases.ORIGINAL_LAB)
    original_lab_address: str = Field(None, alias=ExcelSampleAliases.ORIGINAL_LAB_ADDRESS)
    panels: Annotated[list[str] | None, BeforeValidator(parse_panels)] = Field(
        None, alias=ExcelSampleAliases.PANELS
    )
    pool: str = Field(None, alias=ExcelSampleAliases.POOL)
    post_formalin_fixation_time: str = Field(
        None, alias=ExcelSampleAliases.POST_FORMALIN_FIXATION_TIME
    )
    pre_processing_method: str = Field(None, alias=ExcelSampleAliases.PRE_PROCESSING_METHOD)
    primer: str = Field(None, alias=ExcelSampleAliases.PRIMER)
    priority: Annotated[
        str,
        AfterValidator(convert_to_lower),
        AfterValidator(convert_to_priority),
        AfterValidator(replace_spaces_with_underscores),
    ] = Field(None, alias=ExcelSampleAliases.PRIORITY)
    quantity: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias=ExcelSampleAliases.QUANTITY
    )
    reagent_label: str = Field(None, alias=ExcelSampleAliases.REAGENT_LABEL)
    reference_genome: str = Field(None, alias=ExcelSampleAliases.REFERENCE_GENOME)
    region: str = Field(None, alias=ExcelSampleAliases.REGION)
    region_code: str = Field(None, alias=ExcelSampleAliases.REGION_CODE)
    require_qc_ok: Annotated[bool | None, BeforeValidator(empty_string_to_none)] = Field(
        None, alias=ExcelSampleAliases.REQUIRE_QC_OK
    )
    rml_plate_name: str = Field(None, alias=ExcelSampleAliases.RML_PLATE_NAME)
    selection_criteria: str = Field(None, alias=ExcelSampleAliases.SELECTION_CRITERIA)
    sex: Annotated[str, AfterValidator(convert_sex)] = Field(None, alias=ExcelSampleAliases.SEX)
    source: Annotated[str, AfterValidator(validate_source)] = Field(
        None, alias=ExcelSampleAliases.SOURCE
    )
    status: Annotated[str, AfterValidator(convert_to_lower)] = Field(
        None, alias=ExcelSampleAliases.STATUS
    )
    subject_id: str = Field(None, alias=ExcelSampleAliases.SUBJECT_ID)
    synopsis: str = Field(None, alias=ExcelSampleAliases.SYNOPSIS)
    tissue_block_size: str = Field(None, alias=ExcelSampleAliases.TISSUE_BLOCK_SIZE)
    tumour: Annotated[bool | None, BeforeValidator(empty_string_to_none)] = Field(
        None, alias=ExcelSampleAliases.TUMOUR
    )
    tumour_purity: str = Field(None, alias=ExcelSampleAliases.TUMOUR_PURITY)
    volume: Annotated[str, AfterValidator(numeric_value)] = Field(
        None, alias=ExcelSampleAliases.VOLUME
    )
    well_position: str = Field(None, alias=ExcelSampleAliases.WELL_POSITION)
    well_position_rml: str = Field(None, alias=ExcelSampleAliases.WELL_POSITION_RML)
